from unittest import mock

from django.core import checks
from django.db import connection, models
from django.test import TestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import ArrayField, ObjectIdField
from django_mongodb_backend.indexes import (
    EmbeddedFieldIndex,
    SearchIndex,
    VectorSearchIndex,
)
from django_mongodb_backend.models import EmbeddedModel

from .models import DataHolder, Movie, Owner, Person, Store


@isolate_apps("indexes_")
@mock.patch.object(connection.features, "supports_search", False)
class UnsupportedSearchIndexesTests(TestCase):
    def test_search_requires_search_support(self):
        class Article(models.Model):
            title = models.CharField(max_length=10)

            class Meta:
                indexes = [SearchIndex(fields=["title"])]

        self.assertEqual(
            Article.check(databases={"default"}),
            [
                checks.Warning(
                    "This MongoDB server does not support SearchIndex.",
                    hint=(
                        "The index won't be created. Use a Search-enabled version of MongoDB, "
                        "or silence this warning if you don't care about it."
                    ),
                    obj=Article,
                    id="mongodb.indexes.search.W001",
                )
            ],
        )

    def test_vector_search_requires_search_support(self):
        class Article(models.Model):
            title = models.CharField(max_length=10)
            vector = ArrayField(models.FloatField(), size=10)

            class Meta:
                indexes = [VectorSearchIndex(fields=["title", "vector"], similarities="cosine")]

        self.assertEqual(
            Article.check(databases={"default"}),
            [
                checks.Warning(
                    "This MongoDB server does not support VectorSearchIndex.",
                    hint=(
                        "The index won't be created. Use a Search-enabled version of MongoDB, "
                        "or silence this warning if you don't care about it."
                    ),
                    obj=Article,
                    id="mongodb.indexes.search.W001",
                )
            ],
        )


@isolate_apps("indexes_")
@mock.patch.object(connection.features, "supports_search", True)
class InvalidVectorSearchIndexesTests(TestCase):
    def test_requires_size(self):
        class Article(models.Model):
            title_embedded = ArrayField(models.FloatField())

            class Meta:
                indexes = [VectorSearchIndex(fields=["title_embedded"], similarities="cosine")]

        self.assertEqual(
            Article.check(databases={"default"}),
            [
                checks.Error(
                    "VectorSearchIndex requires 'size' on field 'title_embedded'.",
                    id="mongodb.indexes.search.E002",
                    obj=Article,
                )
            ],
        )

    def test_requires_float_inner_field(self):
        class Article(models.Model):
            title_embedded = ArrayField(models.CharField(max_length=1), size=30)

            class Meta:
                indexes = [VectorSearchIndex(fields=["title_embedded"], similarities="cosine")]

        self.assertEqual(
            Article.check(databases={"default"}),
            [
                checks.Error(
                    "VectorSearchIndex requires the base field of ArrayField "
                    "'title_embedded' to be FloatField or IntegerField but is CharField.",
                    id="mongodb.indexes.search.E003",
                    obj=Article,
                )
            ],
        )

    def test_unsupported_type(self):
        class Article(models.Model):
            data = models.JSONField()
            vector = ArrayField(models.FloatField(), size=10)

            class Meta:
                indexes = [VectorSearchIndex(fields=["data", "vector"], similarities="cosine")]

        self.assertEqual(
            Article.check(databases={"default"}),
            [
                checks.Error(
                    "VectorSearchIndex does not support field 'data' (JSONField).",
                    id="mongodb.indexes.search.E004",
                    obj=Article,
                    hint="Allowed types are boolean, date, number, objectId, string, uuid.",
                )
            ],
        )

    def test_fields_and_similarities_mismatch(self):
        class Article(models.Model):
            vector = ArrayField(models.FloatField(), size=10)

            class Meta:
                indexes = [
                    VectorSearchIndex(
                        fields=["vector"],
                        similarities=["dotProduct", "cosine"],
                    )
                ]

        self.assertEqual(
            Article.check(databases={"default"}),
            [
                checks.Error(
                    "VectorSearchIndex requires the same number of similarities "
                    "and vector fields; Article has 1 ArrayField(s) but similarities "
                    "has 2 element(s).",
                    id="mongodb.indexes.search.E005",
                    obj=Article,
                ),
            ],
        )

    def test_simple(self):
        class Article(models.Model):
            vector = ArrayField(models.FloatField(), size=10)

            class Meta:
                indexes = [VectorSearchIndex(fields=["vector"], similarities="cosine")]

        self.assertEqual(Article.check(databases={"default"}), [])

    def test_valid_fields(self):
        class Data(EmbeddedModel):
            integer = models.IntegerField()

        class SearchIndexTestModel(models.Model):
            text = models.CharField(max_length=100)
            object_id = ObjectIdField()
            number = models.IntegerField()
            vector_integer = ArrayField(models.IntegerField(), size=10)
            vector_float = ArrayField(models.FloatField(), size=10)
            boolean = models.BooleanField()
            date = models.DateTimeField(auto_now=True)

            class Meta:
                indexes = [
                    VectorSearchIndex(
                        name="recent_test_idx",
                        fields=[
                            "text",
                            "object_id",
                            "number",
                            "vector_integer",
                            "vector_float",
                            "boolean",
                            "date",
                        ],
                        similarities="cosine",
                    )
                ]

        self.assertEqual(SearchIndexTestModel.check(databases={"default"}), [])

    def test_requires_vector_field(self):
        class NoSearchVectorModel(models.Model):
            text = models.CharField(max_length=100)

            class Meta:
                indexes = [
                    VectorSearchIndex(
                        name="recent_test_idx", fields=["text"], similarities="cosine"
                    )
                ]

        self.assertEqual(
            NoSearchVectorModel.check(databases={"default"}),
            [
                checks.Error(
                    "VectorSearchIndex requires at least one ArrayField to store vector data.",
                    id="mongodb.indexes.search.E006",
                    obj=NoSearchVectorModel,
                    hint="If you want to perform search operations without vectors, "
                    "use SearchIndex instead.",
                ),
            ],
        )


@isolate_apps("indexes_")
class EmbeddedFieldIndexChecksTests(TestCase):
    def test_valid_embedded_model_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["data.integer"])
        errors = index.check(DataHolder, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_embedded_model_subfield(self):
        index = EmbeddedFieldIndex(
            name="name",
            fields=[
                "data.field_does_not_exist",
                "another_non_existing",
                "data.integer",
            ],
        )
        errors = index.check(DataHolder, connection=connection)
        self.assertEqual(len(errors), 2)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'indexes' refers to the nonexistent field 'data.field_does_not_exist'.",
        )
        self.assertEqual(errors[1].id, "models.E012")
        self.assertEqual(
            errors[1].msg, "'indexes' refers to the nonexistent field 'another_non_existing'."
        )

    def test_valid_embedded_model_array_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["reviews.title"])
        errors = index.check(Movie, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_embedded_model_array_subfield(self):
        index = EmbeddedFieldIndex(
            name="name",
            fields=["reviews.author", "non_existing", "reviews.title"],
        )
        errors = index.check(Movie, connection=connection)
        self.assertEqual(len(errors), 2)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'indexes' refers to the nonexistent field 'reviews.author'.",
        )
        self.assertEqual(errors[1].id, "models.E012")
        self.assertEqual(errors[1].msg, "'indexes' refers to the nonexistent field 'non_existing'.")

    def test_valid_polymorphic_embedded_model_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["pet.name"])
        errors = index.check(Person, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_polymorphic_embedded_model_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["pet.non_existing"])
        errors = index.check(Person, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'indexes' refers to the nonexistent field 'pet.non_existing'.",
        )

    def test_valid_polymorphic_embedded_model_array_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["pets.name"])
        errors = index.check(Owner, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_polymorphic_embedded_model_array_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["pets.non_existing"])
        errors = index.check(Owner, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'indexes' refers to the nonexistent field 'pets.non_existing'.",
        )

    def test_valid_nested_polymorphic_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["thing.tags.name"])
        errors = index.check(Store, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_nested_polymorphic_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["thing.tags.xxx"])
        errors = index.check(Store, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'indexes' refers to the nonexistent field 'thing.tags.xxx'.",
        )



from django.contrib.gis.db.models.lookups import GISLookup
from django.db import NotSupportedError

from django_mongodb_backend.query_utils import process_lhs, process_rhs


def gis_lookup(self, compiler, connection, as_expr=False):
    if as_expr or not self.can_use_path:
        raise NotSupportedError("MongoDB does not support expressions for spatial lookup values.")
    lhs_mql = process_lhs(self, compiler, connection, as_expr=as_expr)
    rhs_mql = process_rhs(self, compiler, connection, as_expr=as_expr)
    try:
        rhs_op = self.get_rhs_op(connection, rhs_mql)
    except KeyError as exc:
        raise NotSupportedError(
            f"MongoDB does not support the '{self.lookup_name}' lookup."
        ) from exc
    return rhs_op.as_mql(lhs_mql, rhs_mql, self.rhs_params)


def register_lookups():
    GISLookup.as_mql = gis_lookup



from django.core import checks
from django.db import NotSupportedError, models

from .managers import EmbeddedModelManager


class EmbeddedModel(models.Model):
    objects = EmbeddedModelManager()

    class Meta:
        abstract = True

    @classmethod
    def check(self, **kwargs):
        errors = super().check(**kwargs)
        for field in self._meta.fields:
            # Ignore auto-created primary keys.
            if field.auto_created:
                continue
            if field.db_index:
                errors.append(
                    checks.Warning(
                        "Using db_index=True on embedded fields is deprecated "
                        "in favor of using EmbeddedFieldIndex in Meta.indexes "
                        "on the top-level model.",
                        obj=field,
                        id="mongodb.fields.embedded_model.W004",
                    )
                )
            if field.unique:
                errors.append(
                    checks.Warning(
                        "Using unique=True on embedded fields is deprecated "
                        "in favor of using EmbeddedFieldUniqueConstraint in "
                        "Meta.constraints on the top-level model.",
                        obj=field,
                        id="mongodb.fields.embedded_model.W005",
                    )
                )
        if self._meta.constraints:
            errors.append(
                checks.Warning(
                    "Using Meta.constraints on embedded models is deprecated "
                    "in favor of using EmbeddedFieldUniqueConstraint in "
                    "Meta.constraints on the top-level model.",
                    obj=self,
                    id="mongodb.fields.embedded_model.W006",
                )
            )
        if self._meta.indexes:
            errors.append(
                checks.Warning(
                    "Using Meta.indexes on embedded models is deprecated in "
                    "favor of using EmbeddedFieldIndex in Meta.indexes on the "
                    "top-level model.",
                    obj=self,
                    id="mongodb.fields.embedded_model.W007",
                )
            )
        if self._meta.unique_together:
            errors.append(
                checks.Warning(
                    "Using Meta.unique_together on embedded models is "
                    "deprecated in favor of using EmbeddedFieldUniqueConstraint "
                    "in Meta.constraints on the top-level model.",
                    obj=self,
                    id="mongodb.fields.embedded_model.W008",
                )
            )
        return errors

    def delete(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be deleted.")

    def save(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be saved.")



import json

from bson import ObjectId, json_util
from django.test import TestCase

from .models import Author


class ExplainTests(TestCase):
    def test_idented(self):
        """The JSON is dumped with indent=4."""
        result = Author.objects.filter().explain()
        self.assertEqual(result[:23], '{\n    "explainVersion":')

    def test_object_id(self):
        """
        The json is dumped with bson.json_util() so that BSON types like ObjectID are
        specially encoded.
        """
        id = ObjectId()
        result = Author.objects.filter(id=id).explain()
        parsed = json_util.loads(result)
        self.assertEqual(parsed["command"]["pipeline"], [{"$match": {"_id": id}}])

    def test_non_ascii(self):
        """The json is dumped with ensure_ascii=False."""
        name = "\U0001d120"
        result = Author.objects.filter(name=name).explain()
        # The non-decoded string must be checked since json.loads() unescapes
        # non-ASCII characters.
        self.assertIn(name, result)
        parsed = json.loads(result)
        self.assertEqual(parsed["command"]["pipeline"], [{"$match": {"name": name}}])



from django.core.exceptions import FieldDoesNotExist
from django.db import connection
from django.test import SimpleTestCase, TestCase

from django_mongodb_backend.indexes import EmbeddedFieldIndex

from .models import DataHolder, Movie, Owner, Person, Store
from .test_base import SchemaAssertionMixin


class EmbeddedFieldIndexNameTests(SimpleTestCase):
    class LongDBTableModel:
        class _meta:
            db_table = "a_really_very_very_long_model_db_table_name"

    def test_long_db_table(self):
        # Long db_table is truncated.
        index = EmbeddedFieldIndex(fields=["some_really_long_field_name"])
        index.set_name_with_model(self.LongDBTableModel)
        self.assertEqual(index.name, "a_really_ve_some_re_6fc5be_idx")

    def test_starting_char(self):
        # Index name must not start with underscore/number even if db_table
        # does.
        class BadModel:
            class _meta:
                db_table = "_1bad_table"

        index = EmbeddedFieldIndex(fields=["field"])
        index.set_name_with_model(BadModel)
        self.assertEqual(index.name, "D1bad_table_field_eb7c16_idx")


class EmbeddedFieldIndexSchemaTests(SchemaAssertionMixin, TestCase):
    def test_embedded_model_subfield(self):
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["data.integer"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=DataHolder)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=DataHolder._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            self.assertEqual(index_info[index.name]["type"], "idx")
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=DataHolder)

    def test_multiple_fields(self):
        index = EmbeddedFieldIndex(
            name="embedded_multi_idx",
            fields=["integer", "data.integer", "data.string"],
        )
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=DataHolder)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=DataHolder._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(
                index_info[index.name]["columns"],
                [*index.fields[:-1], "data.string_"],
            )
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=DataHolder)

    def test_embedded_model_array_subfield(self):
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["reviews.rating"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=Movie)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Movie._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=Movie)

    def test_polymorphic_embedded_model_subfield(self):
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["pet.name"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=Person)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Person._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            self.assertEqual(index_info[index.name]["type"], "idx")
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=Person)

    def test_polymorphic_embedded_model_nonshared_subfield(self):
        """
        Fields that don't exist on all polymorphic embedded models can be
        indexed. In this case, Cat.weight is the second embedded model.
        """
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["pet.weight"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=Person)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Person._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            self.assertEqual(index_info[index.name]["type"], "idx")
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=Person)

    def test_polymorphic_embedded_model_array_subfield(self):
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["pets.name"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=Owner)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Owner._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=Owner)

    def test_polymorphic_embedded_model_aray_nonshared_subfield(self):
        """
        Fields that don't exist on all polymorphic embedded models can be
        indexed. In this case, Cat.weight is the second embedded model.
        """
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["pets.weight"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=Owner)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Owner._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            self.assertEqual(index_info[index.name]["type"], "idx")
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=Owner)

    def test_nested_polymorphic_embedded_array_subfield(self):
        """Traversing PolymorphicEmbeddedModelField + EmbeddModelArrayField"""
        index = EmbeddedFieldIndex(name="embedded_idx", fields=["thing.tags.name"])
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=Store)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Store._meta.db_table,
            )
            self.assertIn(index.name, index_info)
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            self.assertEqual(index_info[index.name]["type"], "idx")
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=Store)


class AddIndexNonexistentFieldTests(TestCase):
    # These cases shouldn't happen as they should be caught by system checks
    # before the migrate stage. Nonetheless, it could be useful to have helpful
    # error messages.

    def test_field(self):
        index = EmbeddedFieldIndex(name="name", fields=["xxx"])
        msg = "Movie has no field named 'xxx"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_index(index=index, model=Movie)

    def test_field_with_dot(self):
        index = EmbeddedFieldIndex(name="name", fields=["title.xxx"])
        msg = "Movie has no field named 'title.xxx"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_index(index=index, model=Movie)

    def test_embedded_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["reviews.xxx"])
        msg = "Review has no field named 'xxx"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_index(index=index, model=Movie)

    def test_polymorphic_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["pet.xxx"])
        msg = "The models of field 'pet' have no field named 'xxx'"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_index(index=index, model=Person)

    def test_nested_polymorphic_embedded_array_subfield(self):
        index = EmbeddedFieldIndex(name="name", fields=["thing.tags.xxx"])
        msg = "The models of field 'thing.tags' have no field named 'xxx'."
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_index(index=index, model=Store)



from django.test import TestCase

from django_mongodb_backend.test import MongoTestCaseMixin

from .models import Author, Book


class DeferTests(MongoTestCaseMixin, TestCase):
    def test_defer(self):
        with self.assertNumQueries(1) as ctx:
            list(Author.objects.defer("name"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__author",
            [{"$project": {"_id": 1}}],
        )

    def test_only(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.only("title"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [{"$project": {"_id": 1, "title": 1}}],
        )



from django.contrib.gis.db.backends.base.features import BaseSpatialFeatures
from django.db import NotSupportedError
from django.utils.functional import cached_property


class GISFeatures(BaseSpatialFeatures):
    has_spatialrefsys_table = False
    supports_distance_geodetic = False
    supports_dwithin_distance_expr = False
    supports_transform = False

    @cached_property
    def django_test_expected_failures(self):
        expected_failures = super().django_test_expected_failures
        expected_failures.update(
            {
                # annotate with Value not supported, e.g.
                # QuerySet.annotate(p=Value(p, GeometryField(srid=4326)
                "gis_tests.geoapp.test_expressions.GeoExpressionsTests.test_geometry_value_annotation",
            }
        )
        return expected_failures

    @cached_property
    def django_test_skips(self):
        skips = super().django_test_skips
        skips.update(
            {
                "inspectdb not supported.": {
                    "gis_tests.inspectapp.tests.InspectDbTests",
                },
                "MongoDB doesn't support the SRID used in this test.": {
                    # Error messages:
                    # - Can't extract geo keys
                    # - Longitude/latitude is out of bounds
                    "gis_tests.geoapp.test_expressions.GeoExpressionsTests.test_update_from_other_field",
                    "gis_tests.layermap.tests.LayerMapTest.test_encoded_name",
                    "gis_tests.relatedapp.tests.RelatedGeoModelTest.test06_f_expressions",
                    # SouthTexasCity fixture objects use SRID 2278 which is ignored
                    # by the patched version of loaddata in the Django fork.
                    "gis_tests.distapp.tests.DistanceTest.test_init",
                },
                "ImproperlyConfigured isn't raised when using RasterField": {
                    # Normally RasterField.db_type() raises an error, but MongoDB
                    # migrations don't need to call it, so the check doesn't happen.
                    "gis_tests.gis_migrations.test_operations.NoRasterSupportTests",
                },
                "GeoJSONSerializer doesn't support ObjectId.": {
                    "gis_tests.geoapp.test_serializers.GeoJSONSerializerTests.test_fields_option",
                    "gis_tests.geoapp.test_serializers.GeoJSONSerializerTests.test_geometry_field_option",
                    "gis_tests.geoapp.test_serializers.GeoJSONSerializerTests.test_serialization_base",
                    "gis_tests.geoapp.test_serializers.GeoJSONSerializerTests.test_srid_option",
                },
            },
        )
        return skips

    django_test_expected_raises = {
        (NotSupportedError, "MongoDB does not support cursor.execute()."): {
            "gis_tests.geoapp.tests.GeoModelTest.test_raw_sql_query",
        },
        (
            NotSupportedError,
            "MongoDB does not support expressions for spatial lookup values.",
        ): {
            "gis_tests.geoapp.tests.GeoLookupTest.test_subquery_annotation",
            "gis_tests.geoapp.tests.GeoQuerySetTest.test_within_subquery",
        },
    }



from datetime import timedelta
from decimal import Decimal

from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import connection, models
from django.test import SimpleTestCase, TestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import PolymorphicEmbeddedModelField
from django_mongodb_backend.models import EmbeddedModel

from .models import Bone, Cat, Dog, Library, Mouse, Person
from .utils import truncate_ms


class MethodTests(SimpleTestCase):
    def test_not_editable(self):
        field = PolymorphicEmbeddedModelField(["Data"], null=True)
        self.assertIs(field.editable, False)

    def test_db_type(self):
        self.assertEqual(PolymorphicEmbeddedModelField(["Data"]).db_type(connection), "object")

    def test_deconstruct(self):
        field = PolymorphicEmbeddedModelField(["Data"], null=True)
        field.name = "field_name"
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(name, "field_name")
        self.assertEqual(path, "django_mongodb_backend.fields.PolymorphicEmbeddedModelField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"embedded_models": ("data",), "null": True})

    def test_deconstruct_model_class(self):
        field = PolymorphicEmbeddedModelField([Cat, Dog])
        *_, kwargs = field.deconstruct()
        self.assertEqual(kwargs["embedded_models"], ("model_fields_.cat", "model_fields_.dog"))

    def test_deconstruct_model_string(self):
        field = PolymorphicEmbeddedModelField(["app_label.Model"])
        *_, kwargs = field.deconstruct()
        self.assertEqual(kwargs["embedded_models"], ("app_label.model",))

    def test_get_db_prep_save_invalid(self):
        msg = (
            "Expected instance of type (<class 'model_fields_.models.Dog'>, "
            "<class 'model_fields_.models.Cat'>), "
            "not <class 'int'>."
        )
        with self.assertRaisesMessage(TypeError, msg):
            Person(pet=42).save()

    def test_validate(self):
        obj = Person(name="Bob", pet=Dog(name="Woofer", barks=None))
        # This isn't quite right because "barks" is the subfield of data
        # that's non-null.
        msg = "{'pet': ['This field cannot be null.']}"
        with self.assertRaisesMessage(ValidationError, msg):
            obj.full_clean()

    def test_validate_wrong_model_type(self):
        obj = Person(name="Bob", pet=Library())
        msg = (
            "{'pet': [\"Expected instance of type "
            "(<class 'model_fields_.models.Dog'>, "
            "<class 'model_fields_.models.Cat'>), not "
            "<class 'model_fields_.models.Library'>.\"]}"
        )
        with self.assertRaisesMessage(ValidationError, msg):
            obj.full_clean()


class ModelTests(TestCase):
    def test_save_load(self):
        Person.objects.create(name="Jim", pet=Dog(name="Woofer"))
        obj = Person.objects.get()
        self.assertIsInstance(obj.pet, Dog)
        # get_prep_value() is called, transforming string to int.
        self.assertEqual(obj.pet.name, "Woofer")
        # Primary keys should not be populated...
        self.assertEqual(obj.pet.id, None)
        # ... unless set explicitly.
        obj.pet.id = obj.id
        obj.save()
        obj = Person.objects.get()
        self.assertEqual(obj.pet.id, obj.id)

    def test_save_load_null(self):
        Person.objects.create(pet=None)
        obj = Person.objects.get()
        self.assertIsNone(obj.pet)

    def test_save_load_decimal(self):
        obj = Person.objects.create(pet=Cat(name="Phoebe", weight="5.5"))
        obj.refresh_from_db()
        self.assertEqual(obj.pet.weight, Decimal("5.5"))

    def test_pre_save(self):
        """Field.pre_save() is called on embedded model fields."""
        obj = Person.objects.create(name="Bob", pet=Dog(name="Woofer"))
        created_at = truncate_ms(obj.pet.created_at)
        updated_at = truncate_ms(obj.pet.updated_at)
        self.assertIsNotNone(obj.pet.created_at)
        # The values may differ by a millisecond since they aren't generated
        # simultaneously.
        self.assertAlmostEqual(updated_at, created_at, delta=timedelta(microseconds=1000))

    def test_missing_field_in_data(self):
        """
        Loading a model with a PolymorphicEmbeddedModelField that has a missing
        subfield (e.g. data not written by Django) that uses a database
        converter (in this case, weight is a DecimalField) doesn't crash.
        """
        Person.objects.create(pet=Cat(name="Pheobe", weight="3.5"))
        connection.database.model_fields__person.update_many({}, {"$unset": {"pet.weight": ""}})
        self.assertIsNone(Person.objects.first().pet.weight)

    def test_embedded_model_field_respects_db_column(self):
        """
        EmbeddedModel data respects Field.db_column. In this case, Cat.name
        has db_column="name_".
        """
        obj = Person.objects.create(pet=Cat(name="Phoebe"))
        query = connection.database.model_fields__person.find({"_id": obj.pk})
        self.assertEqual(query[0]["pet"]["name_"], "Phoebe")


class QueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.cat_owners = [
            Person.objects.create(
                name=f"Cat Owner {x}",
                pet=Cat(
                    name=f"Cat {x}",
                    weight=f"{x}.5",
                    favorite_toy=Mouse(manufacturer=f"Maker {x}"),
                ),
            )
            for x in range(6)
        ]
        cls.dog_owners = [
            Person.objects.create(
                name=f"Dog Owner {x}",
                pet=Dog(
                    name=f"Dog {x}",
                    barks=x % 2 == 0,
                    favorite_toy=Bone(brand=f"Brand {x}"),
                ),
            )
            for x in range(6)
        ]

    def test_exact(self):
        self.assertCountEqual(Person.objects.filter(pet__weight="3.5"), [self.cat_owners[3]])

    def test_lt(self):
        self.assertCountEqual(Person.objects.filter(pet__weight__lt="3.5"), self.cat_owners[:3])

    def test_lte(self):
        self.assertCountEqual(Person.objects.filter(pet__weight__lte="3.5"), self.cat_owners[:4])

    def test_gt(self):
        self.assertCountEqual(Person.objects.filter(pet__weight__gt=3.5), self.cat_owners[4:])

    def test_gte(self):
        self.assertCountEqual(Person.objects.filter(pet__weight__gte=3.5), self.cat_owners[3:])

    def test_range(self):
        self.assertCountEqual(
            Person.objects.filter(pet__weight__range=(2, 4)), self.cat_owners[2:4]
        )

    def test_order_by_embedded_field(self):
        qs = Person.objects.filter(pet__weight__gt=3).order_by("-pet__weight")
        self.assertSequenceEqual(qs, list(reversed(self.cat_owners[3:])))

    def test_boolean(self):
        self.assertCountEqual(
            Person.objects.filter(pet__barks=True),
            [x for i, x in enumerate(self.dog_owners) if i % 2 == 0],
        )

    def test_nested(self):
        # Cat and Dog both have favorite_toy = PolymorphicEmbeddedModelField(...)
        # but with different models. It's possible to query the fields of the
        # Dog's favorite_toy because it's the first model in Person.pet.
        self.assertCountEqual(
            Person.objects.filter(pet__favorite_toy__brand="Brand 1"),
            [self.dog_owners[1]],
        )
        # The fields of Cat can't be queried.
        msg = "The models of field 'favorite_toy' have no field named 'manufacturer'."
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            (Person.objects.filter(pet__favorite_toy__manufacturer="Maker 1"),)


class InvalidLookupTests(SimpleTestCase):
    def test_invalid_field(self):
        msg = "The models of field 'pet' have no field named 'first_name'."
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Person.objects.filter(pet__first_name="Bob")

    def test_invalid_lookup(self):
        msg = "Unsupported lookup 'foo' for CharField 'name'."
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Person.objects.filter(pet__name__foo="Bob")

    def test_invalid_lookup_with_suggestions(self):
        msg = (
            "Unsupported lookup '{lookup}' for CharField 'name', "
            "perhaps you meant {suggested_lookups}?"
        )
        with self.assertRaisesMessage(
            FieldDoesNotExist, msg.format(lookup="exactly", suggested_lookups="exact or iexact")
        ):
            Person.objects.filter(pet__name__exactly="Woof")
        with self.assertRaisesMessage(
            FieldDoesNotExist, msg.format(lookup="gti", suggested_lookups="gt or gte")
        ):
            Person.objects.filter(pet__name__gti="Woof")
        with self.assertRaisesMessage(
            FieldDoesNotExist, msg.format(lookup="is_null", suggested_lookups="isnull")
        ):
            Person.objects.filter(pet__name__is_null="Woof")


@isolate_apps("model_fields_")
class CheckTests(SimpleTestCase):
    def test_no_relational_fields(self):
        class Target(EmbeddedModel):
            key = models.ForeignKey("MyModel", models.CASCADE)

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelField([Target])

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.embedded_model.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg, "Embedded models cannot have relational fields (Target.key is a ForeignKey)."
        )

    def test_embedded_model_subclass(self):
        class Target(models.Model):
            pass

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelField([Target])

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.embedded_model.E002")
        msg = errors[0].msg
        self.assertEqual(
            msg,
            "Embedded models must be a subclass of django_mongodb_backend.models.EmbeddedModel.",
        )

    def test_clashing_fields(self):
        class Target1(EmbeddedModel):
            clash = models.DecimalField(max_digits=4, decimal_places=2)

        class Target2(EmbeddedModel):
            clash = models.CharField(max_length=255)

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelField([Target1, Target2])

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.embedded_model.W003")
        self.assertEqual(
            errors[0].msg,
            "Embedded models model_fields_.Target1 and model_fields_.Target2 "
            "both have field 'clash' of different type.",
        )
        self.assertEqual(
            errors[0].hint,
            "It may be impossible to query both fields.",
        )

    def test_clashing_fields_of_same_type(self):
        """Fields of different type don't clash if they use the same db_type."""

        class Target1(EmbeddedModel):
            clash = models.TextField()

        class Target2(EmbeddedModel):
            clash = models.CharField(max_length=255)

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelField([Target1, Target2])

        errors = MyModel().check()
        self.assertEqual(len(errors), 0)



from django.db import connections


def get_mongodb_connection():
    for alias in connections:
        if connections[alias].vendor == "mongodb":
            return connections[alias]
    return None


def serialize_model_reference(model):
    """Serialize a model to its label for use in migrations."""
    if isinstance(model, str):
        # For "app_label.Model", lowercase the model name only.
        if "." in model:
            app_label, model_name = model.split(".")
            return f"{app_label}.{model_name.lower()}"
        # For "Model", lowercase it.
        return model.lower()
    return model._meta.label_lower



from django.core.exceptions import ImproperlyConfigured
from django.db import models

from django_mongodb_backend.fields import ArrayField, EmbeddedModelArrayField, EmbeddedModelField
from django_mongodb_backend.fields.objectid import ObjectIdField

from .mixins import NoEncryptedEmbeddedFieldsMixin


class EncryptedFieldMixin:
    encrypted = True

    def __init__(self, *args, queries=None, db_index=False, null=False, unique=False, **kwargs):
        if db_index:
            raise ValueError("Encrypted fields do not support db_index=True.")
        if null:
            raise ValueError("Encrypted fields do not support null=True.")
        if unique:
            raise ValueError("Encrypted fields do not support unique=True.")
        self.queries = queries
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.queries is not None:
            kwargs["queries"] = self.queries
        if path.startswith("django_mongodb_backend.fields.encryption"):
            path = path.replace(
                "django_mongodb_backend.fields.encryption",
                "django_mongodb_backend.fields",
            )
        return name, path, args, kwargs

    def get_db_prep_save(self, value, connection):
        if not connection.auto_encryption_opts:
            raise ImproperlyConfigured(
                f"Cannot save encrypted field '{self.name}' in non-encrypted "
                f"database '{connection.alias}'."
            )
        return super().get_db_prep_save(value, connection)


class NoQueriesMixin:
    def __init__(self, *args, **kwargs):
        if "queries" in kwargs:
            raise ValueError(f"{self.__class__.__name__} does not support the queries argument.")
        super().__init__(*args, **kwargs)


# Django fields
class EncryptedBinaryField(EncryptedFieldMixin, models.BinaryField):
    pass


class EncryptedBigIntegerField(EncryptedFieldMixin, models.BigIntegerField):
    pass


class EncryptedBooleanField(EncryptedFieldMixin, models.BooleanField):
    pass


class EncryptedCharField(EncryptedFieldMixin, models.CharField):
    pass


class EncryptedDateField(EncryptedFieldMixin, models.DateField):
    pass


class EncryptedDateTimeField(EncryptedFieldMixin, models.DateTimeField):
    pass


class EncryptedDecimalField(EncryptedFieldMixin, models.DecimalField):
    pass


class EncryptedDurationField(EncryptedFieldMixin, models.DurationField):
    pass


class EncryptedEmailField(EncryptedFieldMixin, models.EmailField):
    pass


class EncryptedFloatField(EncryptedFieldMixin, models.FloatField):
    pass


class EncryptedGenericIPAddressField(EncryptedFieldMixin, models.GenericIPAddressField):
    pass


class EncryptedIntegerField(EncryptedFieldMixin, models.IntegerField):
    pass


class EncryptedPositiveBigIntegerField(EncryptedFieldMixin, models.PositiveBigIntegerField):
    pass


class EncryptedPositiveIntegerField(EncryptedFieldMixin, models.PositiveIntegerField):
    pass


class EncryptedPositiveSmallIntegerField(EncryptedFieldMixin, models.PositiveSmallIntegerField):
    pass


class EncryptedSmallIntegerField(EncryptedFieldMixin, models.SmallIntegerField):
    pass


class EncryptedTextField(EncryptedFieldMixin, models.TextField):
    pass


class EncryptedTimeField(EncryptedFieldMixin, models.TimeField):
    pass


class EncryptedURLField(EncryptedFieldMixin, models.URLField):
    pass


class EncryptedUUIDField(EncryptedFieldMixin, models.UUIDField):
    pass


# MongoDB fields
class EncryptedArrayField(NoQueriesMixin, EncryptedFieldMixin, ArrayField):
    pass


class EncryptedEmbeddedModelArrayField(
    NoQueriesMixin, EncryptedFieldMixin, EmbeddedModelArrayField
):
    pass


class EncryptedEmbeddedModelField(
    NoEncryptedEmbeddedFieldsMixin, NoQueriesMixin, EncryptedFieldMixin, EmbeddedModelField
):
    pass


class EncryptedObjectIdField(EncryptedFieldMixin, ObjectIdField):
    pass



from django.core.exceptions import FieldDoesNotExist
from django.db import IntegrityError, connection
from django.test import TestCase

from django_mongodb_backend.constraints import EmbeddedFieldUniqueConstraint

from .models import Address, Cat, Data, DataHolder, Dog, Movie, Owner, Person, Review, Store, Tag


class EmbeddedFieldUniqueConstraintTests(TestCase):
    def test_embedded_model_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_created_constraint",
            fields=["data.integer"],
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=DataHolder)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=DataHolder._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            self.assertEqual(constraint_info[constraint.name]["type"], "idx")
            DataHolder.objects.create(data=Data(integer=1))
            msg = "embedded_created_constraint dup key: { data.integer: 1 }"
            with self.assertRaisesMessage(IntegrityError, msg):
                DataHolder.objects.create(data=Data(integer=1))
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=DataHolder)

    def test_multiple_fields(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_multi_constraint",
            fields=["integer", "data.integer"],
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=DataHolder)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=DataHolder._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=DataHolder)

    def test_embedded_model_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_multi_idx",
            fields=["reviews.rating"],
            nulls_distinct=False,
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=Movie)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Movie._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            Movie.objects.create(title="Jaws", reviews=[Review(title="Good", rating=1)])
            msg = "embedded_multi_idx dup key: { reviews.rating: 1 }"
            with self.assertRaisesMessage(IntegrityError, msg):
                Movie.objects.create(
                    title="Jaws",
                    reviews=[Review(title="Good", rating=2), Review(title="Bad", rating=1)],
                )
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=Movie)

    def test_polymorphic_embedded_model_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_created_constraint",
            fields=["pet.name"],
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=Person)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Person._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            self.assertEqual(constraint_info[constraint.name]["type"], "idx")
            Person.objects.create(pet=Dog(name="Woofer"))
            msg = 'embedded_created_constraint dup key: { pet.name: "Woofer"'
            with self.assertRaisesMessage(IntegrityError, msg):
                Person.objects.create(pet=Dog(name="Woofer"))
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=Person)

    def test_polymorphic_embedded_model_nonshared_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_created_constraint",
            fields=["pet.weight"],
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=Person)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Person._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            self.assertEqual(constraint_info[constraint.name]["type"], "idx")
            Person.objects.create(pet=Cat(name="Pheobe", weight="10"))
            msg = "embedded_created_constraint dup key: { pet.weight: 10 }"
            with self.assertRaisesMessage(IntegrityError, msg):
                Person.objects.create(pet=Cat(name="Pheobe", weight="10"))
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=Person)

    def test_polymorphic_embedded_model_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_created_constraint",
            fields=["pets.name"],
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=Owner)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Owner._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            self.assertEqual(constraint_info[constraint.name]["type"], "idx")
            Owner.objects.create(pets=[Dog(name="Woofer")])
            msg = 'embedded_created_constraint dup key: { pets.name: "Woofer"'
            with self.assertRaisesMessage(IntegrityError, msg):
                Owner.objects.create(pets=[Dog(name="Woofer")])
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=Owner)

    def test_polymorphic_embedded_model_array_nonshared_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="embedded_created_constraint",
            fields=["pets.weight"],
        )
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=Owner)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Owner._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            self.assertEqual(constraint_info[constraint.name]["type"], "idx")
            Owner.objects.create(pets=[Cat(name="Pheobe", weight="10")])
            msg = "embedded_created_constraint dup key: { pets.weight: 10 }"
            with self.assertRaisesMessage(IntegrityError, msg):
                Owner.objects.create(pets=[Cat(name="Pheobe", weight="10")])
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=Owner)

    def test_nested_polymorphic_embedded_array_subfield(self):
        """Traversing PolymorphicEmbeddedModelField + EmbeddModelArrayField"""
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["thing.tags.name"])
        with connection.schema_editor() as editor:
            editor.add_constraint(constraint=constraint, model=Store)
        try:
            constraint_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=Store._meta.db_table,
            )
            self.assertIn(constraint.name, constraint_info)
            self.assertCountEqual(
                constraint_info[constraint.name]["columns"],
                constraint.fields,
            )
            self.assertEqual(constraint_info[constraint.name]["type"], "idx")
            Store.objects.create(thing=Address(tags=[Tag(name="dup")]))
            msg = 'name dup key: { thing.tags.name: "dup" }'
            with self.assertRaisesMessage(IntegrityError, msg):
                Store.objects.create(thing=Address(tags=[Tag(name="dup")]))
        finally:
            with connection.schema_editor() as editor:
                editor.remove_constraint(constraint=constraint, model=Store)


class AddConstraintNonexistentFieldTests(TestCase):
    # These cases shouldn't happen as they should be caught by system checks
    # before the migrate stage. Nonetheless, it could be useful to have helpful
    # error messages.

    def test_field(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["xxx"])
        msg = "Movie has no field named 'xxx"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_constraint(constraint=constraint, model=Movie)

    def test_field_with_dot(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["title.xxx"])
        msg = "Movie has no field named 'title.xxx"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_constraint(constraint=constraint, model=Movie)

    def test_embedded_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["reviews.xxx"])
        msg = "Review has no field named 'xxx"
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_constraint(constraint=constraint, model=Movie)

    def test_polymorphic_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["pet.xxx"])
        msg = "The models of field 'pet' have no field named 'xxx'."
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_constraint(constraint=constraint, model=Person)

    def test_nested_polymorphic_embedded_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["thing.tags.xxx"])
        msg = "The models of field 'thing.tags' have no field named 'xxx'."
        with connection.schema_editor() as editor, self.assertRaisesMessage(FieldDoesNotExist, msg):
            editor.add_constraint(constraint=constraint, model=Store)



from functools import reduce

import pymongo
from bson.binary import Binary
from django.conf import settings
from django.db import connections
from django.test import TestCase, skipUnlessDBFeature


@skipUnlessDBFeature("supports_queryable_encryption")
class EncryptionTestCase(TestCase):
    databases = {"default", "encrypted"}
    maxDiff = None

    def assertEncrypted(self, model, field_name):
        """
        Assert that the data in a give model's field contains binary data (and
        thus is probably encrypted).
        """
        # Create an unencrypted connection to the encrypted database.
        conn_params = connections["default"].get_connection_params()
        db_name = settings.DATABASES["encrypted"]["NAME"]
        with pymongo.MongoClient(**conn_params) as new_connection:
            db = new_connection[db_name]
            collection = db[model._meta.db_table]
            data = collection.find_one({}, {field_name: 1, "_id": 0})
            # Get the value for embedded fields (separated by dots), if
            # applicable.
            field_value = reduce(dict.__getitem__, field_name.split("."), data)
            self.assertIsInstance(field_value, Binary)



from decimal import Decimal

from django.core.exceptions import FieldDoesNotExist
from django.db import connection, models
from django.test import SimpleTestCase, TestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import PolymorphicEmbeddedModelArrayField
from django_mongodb_backend.models import EmbeddedModel

from .models import Bone, Cat, Dog, Owner


class MethodTests(SimpleTestCase):
    def test_not_editable(self):
        field = PolymorphicEmbeddedModelArrayField(["Dog"])
        self.assertIs(field.editable, False)

    def test_deconstruct(self):
        field = PolymorphicEmbeddedModelArrayField(["Dog"], null=True)
        field.name = "field_name"
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(name, "field_name")
        self.assertEqual(path, "django_mongodb_backend.fields.PolymorphicEmbeddedModelArrayField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"embedded_models": ("dog",), "null": True})

    def test_deconstruct_model_class(self):
        field = PolymorphicEmbeddedModelArrayField([Cat, Dog])
        *_, kwargs = field.deconstruct()
        self.assertEqual(kwargs["embedded_models"], ("model_fields_.cat", "model_fields_.dog"))

    def test_deconstruct_model_string(self):
        field = PolymorphicEmbeddedModelArrayField(["app_label.Model"])
        *_, kwargs = field.deconstruct()
        self.assertEqual(kwargs["embedded_models"], ("app_label.model",))

    def test_size_not_supported(self):
        msg = "PolymorphicEmbeddedModelArrayField does not support size."
        with self.assertRaisesMessage(ValueError, msg):
            PolymorphicEmbeddedModelArrayField("Data", size=1)

    def test_get_db_prep_save_invalid(self):
        msg = (
            "Expected list of (<class 'model_fields_.models.Dog'>, "
            "<class 'model_fields_.models.Cat'>) instances, not <class 'int'>."
        )
        with self.assertRaisesMessage(TypeError, msg):
            Owner(pets=42).save()

    def test_get_db_prep_save_invalid_list(self):
        msg = (
            "Expected instance of type (<class 'model_fields_.models.Dog'>, "
            "<class 'model_fields_.models.Cat'>), not <class 'int'>."
        )
        with self.assertRaisesMessage(TypeError, msg):
            Owner(pets=[42]).save()


class ModelTests(TestCase):
    def test_save_load(self):
        pets = [Dog(name="Woofer"), Cat(name="Phoebe", weight="3.5")]
        Owner.objects.create(name="Bob", pets=pets)
        owner = Owner.objects.get(name="Bob")
        self.assertEqual(owner.pets[0].name, "Woofer")
        self.assertEqual(owner.pets[1].name, "Phoebe")
        self.assertEqual(owner.pets[1].weight, Decimal("3.5"))
        self.assertEqual(len(owner.pets), 2)

    def test_save_load_null(self):
        Owner.objects.create(name="Bob")
        owner = Owner.objects.get(name="Bob")
        self.assertIsNone(owner.pets)

    def test_missing_field_in_data(self):
        """
        Loading a model with a PolymorphicEmbeddedModelArrayField that has a
        missing subfield (e.g. data not written by Django) that uses a database
        converter (in this case, weight is a DecimalField) doesn't crash.
        """
        Owner.objects.create(name="Bob", pets=[Cat(name="Phoebe", weight="3.5")])
        connection.database.model_fields__owner.update_many({}, {"$unset": {"pets.$[].weight": ""}})
        self.assertIsNone(Owner.objects.first().pets[0].weight)

    def test_embedded_model_field_respects_db_column(self):
        """
        EmbeddedModel data respects Field.db_column. In this case, Cat.name
        has db_column="name_".
        """
        obj = Owner.objects.create(name="Bob", pets=[Cat(name="Phoebe", weight="3.5")])
        query = connection.database.model_fields__owner.find({"_id": obj.pk})
        self.assertEqual(query[0]["pets"][0]["name_"], "Phoebe")


class QueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.fred = Owner.objects.create(
            name="Fred",
            pets=[
                Dog(name="Woofer", toys=[Bone(brand="Brand 1")]),
                Cat(name="Phoebe", weight="3.5"),
            ],
        )
        cls.bob = Owner.objects.create(
            name="Bob",
            pets=[Dog(name="Lassy", toys=[Bone(brand="Brand 1")])],
        )
        cls.mary = Owner.objects.create(
            name="Mary",
            pets=[
                Dog(name="Doodle"),
                Cat(name="Tyler"),
            ],
        )
        cls.julie = Owner.objects.create(
            name="Mary",
            pets=[
                Cat(name="General"),
                Dog(name="Skip", toys=[Bone(brand="Brand 2")]),
            ],
        )

    def test_exact(self):
        self.assertCountEqual(Owner.objects.filter(pets__name="Woofer"), [self.fred])

    def test_array_index(self):
        self.assertCountEqual(Owner.objects.filter(pets__0__name="Lassy"), [self.bob])

    def test_nested_array_index(self):
        self.assertCountEqual(
            Owner.objects.filter(pets__toys__0__brand="Brand 1"), [self.fred, self.bob]
        )

    def test_array_slice(self):
        self.assertSequenceEqual(Owner.objects.filter(pets__0_1__name="Woofer"), [self.fred])

    def test_len(self):
        self.assertCountEqual(Owner.objects.filter(pets__len=3), [])
        self.assertCountEqual(
            Owner.objects.filter(pets__len=2),
            [self.fred, self.mary, self.julie],
        )
        # Nested PolymorphicEmbeddedModelField
        self.assertCountEqual(
            Owner.objects.filter(pets__toys__len=1), [self.fred, self.bob, self.julie]
        )
        self.assertCountEqual(Owner.objects.filter(pets__toys__len=2), [])
        # Nested Indexed Array
        self.assertCountEqual(Owner.objects.filter(pets__0__toys__len=1), [self.fred, self.bob])
        self.assertCountEqual(Owner.objects.filter(pets__0__toys__len=0), [])
        self.assertCountEqual(Owner.objects.filter(pets__1__toys__len=1), [self.julie])

    def test_in(self):
        self.assertCountEqual(Owner.objects.filter(pets__weight__in=["4.0"]), [])
        self.assertCountEqual(Owner.objects.filter(pets__weight__in=["3.5"]), [self.fred])

    def test_iexact(self):
        self.assertCountEqual(Owner.objects.filter(pets__name__iexact="woofer"), [self.fred])

    def test_gt(self):
        self.assertCountEqual(Owner.objects.filter(pets__weight__gt=1), [self.fred])

    def test_gte(self):
        self.assertCountEqual(Owner.objects.filter(pets__weight__gte=1), [self.fred])

    def test_lt(self):
        self.assertCountEqual(Owner.objects.filter(pets__weight__lt=2), [])

    def test_lte(self):
        self.assertCountEqual(Owner.objects.filter(pets__weight__lte=2), [])

    def test_querying_array_not_allowed(self):
        msg = (
            "Lookups aren't supported on PolymorphicEmbeddedModelArrayField. "
            "Try querying one of its embedded fields instead."
        )
        with self.assertRaisesMessage(ValueError, msg):
            Owner.objects.filter(pets=10).first()

        with self.assertRaisesMessage(ValueError, msg):
            Owner.objects.filter(pets__0_1=10).first()

    def test_invalid_field(self):
        msg = "The models of field 'pets' have no field named 'xxx'."
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Owner.objects.filter(pets__xxx=10).first()

    def test_invalid_nested_field(self):
        msg = "The models of field 'toys' have no field named 'xxx'."
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Owner.objects.filter(pets__toys__xxx=10).first()

    def test_invalid_lookup(self):
        msg = "Unsupported lookup 'return' for PolymorphicEmbeddedModelArrayField of 'CharField'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Owner.objects.filter(pets__name__return="xxx")

    def test_unsupported_lookup(self):
        msg = "Unsupported lookup 'range' for PolymorphicEmbeddedModelArrayField of 'DecimalField'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Owner.objects.filter(pets__weight__range=[10])

    def test_missing_lookup_suggestions(self):
        msg = (
            "Unsupported lookup 'ltee' for PolymorphicEmbeddedModelArrayField "
            "of 'DecimalField', perhaps you meant lte or lt?"
        )
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Owner.objects.filter(pets__weight__ltee=3)

    def test_nested_lookup(self):
        msg = "Cannot perform multiple levels of array traversal in a query."
        with self.assertRaisesMessage(ValueError, msg):
            Owner.objects.filter(pets__toys__brand="")


@isolate_apps("model_fields_")
class CheckTests(SimpleTestCase):
    def test_no_relational_fields(self):
        class Target(EmbeddedModel):
            key = models.ForeignKey("MyModel", models.CASCADE)

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelArrayField([Target])

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.array.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg,
            "Base field for array has errors:\n    "
            "Embedded models cannot have relational fields (Target.key is a ForeignKey). "
            "(mongodb.fields.embedded_model.E001)",
        )

    def test_embedded_model_subclass(self):
        class Target(models.Model):
            pass

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelArrayField([Target])

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.array.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg,
            "Base field for array has errors:\n    "
            "Embedded models must be a subclass of "
            "django_mongodb_backend.models.EmbeddedModel. "
            "(mongodb.fields.embedded_model.E002)",
        )

    def test_clashing_fields(self):
        class Target1(EmbeddedModel):
            clash = models.DecimalField(max_digits=4, decimal_places=2)

        class Target2(EmbeddedModel):
            clash = models.CharField(max_length=255)

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelArrayField([Target1, Target2])

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.array.W004")
        self.assertEqual(
            errors[0].msg,
            "Base field for array has warnings:\n    "
            "Embedded models model_fields_.Target1 and model_fields_.Target2 "
            "both have field 'clash' of different type. "
            "(mongodb.fields.embedded_model.W003)",
        )

    def test_clashing_fields_of_same_type(self):
        """Fields of different type don't clash if they use the same db_type."""

        class Target1(EmbeddedModel):
            clash = models.TextField()

        class Target2(EmbeddedModel):
            clash = models.CharField(max_length=255)

        class MyModel(models.Model):
            field = PolymorphicEmbeddedModelArrayField([Target1, Target2])

        errors = MyModel().check()
        self.assertEqual(len(errors), 0)



import contextlib

from django.apps import apps
from django.core import checks
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.db.models.fields.related import lazy_related_operation
from django.utils.functional import cached_property

from .embedded_model import EmbeddedModelTransformFactory
from .utils import get_mongodb_connection, serialize_model_reference


class PolymorphicEmbeddedModelField(models.Field):
    """Field that stores a model instance of varying type."""

    stores_model_instance = True

    def __init__(self, embedded_models, *args, **kwargs):
        """
        `embedded_models` is a list of possible model classes to be stored.
        Like other relational fields, each model may also be passed as a
        string.
        """
        self.embedded_models = embedded_models
        kwargs["editable"] = False
        super().__init__(*args, **kwargs)

    def db_type(self, connection):
        return "object"

    def check(self, **kwargs):
        from django_mongodb_backend.models import EmbeddedModel  # noqa: PLC0415

        errors = super().check(**kwargs)
        embedded_fields = {}
        for model in self.embedded_models:
            if not issubclass(model, EmbeddedModel):
                return [
                    checks.Error(
                        "Embedded models must be a subclass of "
                        "django_mongodb_backend.models.EmbeddedModel.",
                        obj=self,
                        hint="{model} doesn't subclass EmbeddedModel.",
                        id="mongodb.fields.embedded_model.E002",
                    )
                ]
            for field in model._meta.fields:
                if field.remote_field:
                    errors.append(
                        checks.Error(
                            "Embedded models cannot have relational fields "
                            f"({model().__class__.__name__}.{field.name} "
                            f"is a {field.__class__.__name__}).",
                            obj=self,
                            id="mongodb.fields.embedded_model.E001",
                        )
                    )
                field_name = field.name
                if existing_field := embedded_fields.get(field.name):
                    connection = get_mongodb_connection()
                    if existing_field.db_type(connection) != field.db_type(connection):
                        errors.append(
                            checks.Warning(
                                f"Embedded models {existing_field.model._meta.label} "
                                f"and {field.model._meta.label} both have field "
                                f"'{field_name}' of different type.",
                                obj=self,
                                id="mongodb.fields.embedded_model.W003",
                                hint="It may be impossible to query both fields.",
                            )
                        )

                else:
                    embedded_fields[field_name] = field
        return errors

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("django_mongodb_backend.fields.polymorphic_embedded_model"):
            path = path.replace(
                "django_mongodb_backend.fields.polymorphic_embedded_model",
                "django_mongodb_backend.fields",
            )
        kwargs["embedded_models"] = tuple(
            serialize_model_reference(x) for x in self.embedded_models
        )
        del kwargs["editable"]
        return name, path, args, kwargs

    def get_internal_type(self):
        return "PolymorphicEmbeddedModelField"

    def _set_model(self, model):
        """
        Resolve embedded model classes once the field knows the model it
        belongs to. If any of the items in __init__()'s embedded_models
        argument are strings, resolve each to the actual model class, similar
        to relational fields.
        """
        self._model = model
        if model is not None:
            for embedded_model in self.embedded_models:
                if isinstance(embedded_model, str):

                    def _resolve_lookup(_, *resolved_models):
                        self.embedded_models = resolved_models

                    lazy_related_operation(_resolve_lookup, model, *self.embedded_models)

    model = property(lambda self: self._model, _set_model)

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def to_python(self, value):
        """
        Pass embedded model fields' values through each field's to_python() and
        reinstantiate the embedded instance.
        """
        if value is None:
            return None
        if not isinstance(value, dict):
            return value
        model_class = self._get_model_from_label(value.pop("_label"))
        instance = model_class(
            **{
                field.attname: field.to_python(value[field.column])
                for field in model_class._meta.fields
                if field.column in value
            }
        )
        instance._state.adding = False
        return instance

    def get_db_prep_save(self, embedded_instance, connection):
        """
        Apply pre_save() and get_db_prep_save() of embedded instance fields and
        create the {field: value} dict to be saved.
        """
        if embedded_instance is None:
            return None
        if not isinstance(embedded_instance, self.embedded_models):
            raise TypeError(
                f"Expected instance of type {self.embedded_models!r}, not "
                f"{type(embedded_instance)!r}."
            )
        field_values = {}
        add = embedded_instance._state.adding
        for field in embedded_instance._meta.fields:
            value = field.get_db_prep_save(
                field.pre_save(embedded_instance, add), connection=connection
            )
            # Exclude unset primary keys (e.g. {'id': None}).
            if field.primary_key and value is None:
                continue
            field_values[field.column] = value
        # Store the model's label to know the class to use for initializing
        # upon retrieval.
        field_values["_label"] = embedded_instance._meta.label
        # This instance will exist in the database soon.
        embedded_instance._state.adding = False
        return field_values

    @cached_property
    def resolved_embedded_models(self):
        # Since Field.deconstruct() serializes self.embedded_models as a list
        # of strings, these strings may need to be resolved after a
        # Field.clone() in querying.
        return tuple(
            apps.get_model(model) if isinstance(model, str) else model
            for model in self.embedded_models
        )

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        for model in self.resolved_embedded_models:
            with contextlib.suppress(FieldDoesNotExist):
                field = model._meta.get_field(name)
                break
        else:
            raise FieldDoesNotExist(
                f"The models of field '{self.name}' have no field named '{name}'."
            )
        return EmbeddedModelTransformFactory(field)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        if not isinstance(value, self.embedded_models):
            raise ValidationError(
                f"Expected instance of type {self.embedded_models!r}, not {type(value)!r}."
            )
        for field in value._meta.fields:
            attname = field.attname
            field.validate(getattr(value, attname), model_instance)

    def formfield(self, **kwargs):
        raise NotImplementedError("PolymorphicEmbeddedModelField does not support forms.")

    def _get_model_from_label(self, label):
        return next(model for model in self.embedded_models if model._meta.label == label)



from django.db import IntegrityError
from django.test import TestCase

from .models import UniqueNumber


class UpdateTests(TestCase):
    def test_integrity_error(self):
        UniqueNumber.objects.create(number=1)
        UniqueNumber.objects.create(number=2)
        msg = "duplicate key error collection"
        with self.assertRaisesMessage(IntegrityError, msg):
            UniqueNumber.objects.filter(number=1).update(number=2)



from datetime import timedelta

from django.db import IntegrityError
from django.test import TestCase

from .models import UniqueIntegers


class UniqueTests(TestCase):
    def test_small_value(self):
        """
        Duplicate values < 32 bits are prohibited. This confirms DurationField
        values are cast to Int64 so MongoDB stores them as long. Otherwise, the
        partialFilterExpression: {$type: long} unique constraint doesn't work.
        """
        UniqueIntegers.objects.create(duration=timedelta(1))
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(duration=timedelta(1))

    def test_large_value(self):
        """
        Duplicate values > 32 bits are prohibited. This confirms DurationField
        uses the long db_type() rather than the 32 bit int type.
        """
        UniqueIntegers.objects.create(duration=timedelta(1000000))
        with self.assertRaises(IntegrityError):
            UniqueIntegers.objects.create(duration=timedelta(1000000))



"""Mixins used across multiple modules (to avoid circular imports)."""

from django.core.checks import Error


class NoEncryptedEmbeddedFieldsMixin:
    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_no_encrypted_fields(),
        ]

    def _check_no_encrypted_fields(self):
        for field in self.embedded_model._meta.fields:
            if getattr(field, "encrypted", False):
                return [
                    Error(
                        f"{self.__class__.__name__} cannot contain encrypted "
                        f"fields (found {field.__class__.__name__}).",
                        obj=self,
                        id="mongodb.fields.encryption.E001",
                    )
                ]
        return []



import difflib

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Field, lookups
from django.db.models.expressions import Col
from django.db.models.fields.related import lazy_related_operation
from django.db.models.lookups import Lookup, Transform
from django.utils.functional import cached_property

from django_mongodb_backend import forms
from django_mongodb_backend.fields import EmbeddedModelField
from django_mongodb_backend.fields.array import ArrayField, ArrayLenTransform
from django_mongodb_backend.query_utils import process_lhs, process_rhs

from .mixins import NoEncryptedEmbeddedFieldsMixin
from .utils import serialize_model_reference


class EmbeddedModelArrayField(NoEncryptedEmbeddedFieldsMixin, ArrayField):
    def __init__(self, embedded_model, **kwargs):
        if "size" in kwargs:
            raise ValueError("EmbeddedModelArrayField does not support size.")
        super().__init__(EmbeddedModelField(embedded_model), **kwargs)
        self.embedded_model = embedded_model

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        super().contribute_to_class(cls, name, private_only=private_only, **kwargs)

        if not cls._meta.abstract:
            # If embedded_models contains any strings, replace them with the actual
            # model classes.
            def _resolve_lookup(_, resolved_model):
                self.embedded_model = resolved_model

            lazy_related_operation(_resolve_lookup, cls, self.embedded_model)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path == "django_mongodb_backend.fields.embedded_model_array.EmbeddedModelArrayField":
            path = "django_mongodb_backend.fields.EmbeddedModelArrayField"
        kwargs["embedded_model"] = serialize_model_reference(self.embedded_model)
        del kwargs["base_field"]
        return name, path, args, kwargs

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, (list, tuple)):
            # Must call get_db_prep_save() rather than get_db_prep_value()
            # to transform model instances to dicts.
            return [self.base_field.get_db_prep_save(i, connection) for i in value]
        if value is not None:
            raise TypeError(
                f"Expected list of {self.embedded_model!r} instances, not {type(value)!r}."
            )
        return value

    def formfield(self, **kwargs):
        # Skip ArrayField.formfield() which has some differences, including
        # unneeded "base_field", and "max_length" instead of "max_num".
        return Field.formfield(
            self,
            **{
                "form_class": forms.EmbeddedModelArrayField,
                "model": self.embedded_model,
                "max_num": self.max_size,
                "prefix": self.name,
                **kwargs,
            },
        )

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        field = self.base_field.resolved_embedded_model._meta.get_field(name)
        return EmbeddedModelArrayFieldTransformFactory(field)

    def _get_lookup(self, lookup_name):
        lookup = super()._get_lookup(lookup_name)
        if lookup is None or lookup is ArrayLenTransform:
            return lookup

        class EmbeddedModelArrayFieldLookups(Lookup):
            def as_mql(self, compiler, connection, as_expr=False):
                raise ValueError(
                    "Lookups aren't supported on EmbeddedModelArrayField. "
                    "Try querying one of its embedded fields instead."
                )

        return EmbeddedModelArrayFieldLookups


class _EmbeddedModelArrayOutputField(ArrayField):
    """
    Represent the output of an EmbeddedModelArrayField when traversed in a
    query path.

    This field is not meant to be used in model definitions. It exists solely
    to support query output resolution. When an EmbeddedModelArrayField is
    accessed in a query, the result should behave like an array of the embedded
    model's target type.

    While it mimics ArrayField's lookup behavior, the way those lookups are
    resolved follows the semantics of EmbeddedModelArrayField rather than
    ArrayField.
    """

    ALLOWED_LOOKUPS = {
        "in",
        "exact",
        "iexact",
        "gt",
        "gte",
        "lt",
        "lte",
    }

    def get_lookup(self, name):
        return super().get_lookup(name) if name in self.ALLOWED_LOOKUPS else None


class EmbeddedModelArrayFieldBuiltinLookup(Lookup):
    def process_rhs(self, compiler, connection, as_expr=False):
        value = self.rhs
        if not self.get_db_prep_lookup_value_is_iterable:
            value = [value]
        # Value must be serialized based on the query target. If querying a
        # subfield inside the array (i.e., a nested KeyTransform), use the
        # output field of the subfield. Otherwise, use the base field of the
        # array itself.
        get_db_prep_value = self.lhs._lhs.output_field.get_db_prep_value
        return None, [
            v if hasattr(v, "as_mql") else get_db_prep_value(v, connection, prepared=True)
            for v in value
        ]

    def as_mql_expr(self, compiler, connection):
        # Querying a subfield within the array elements (via nested
        # KeyTransform). Replicate MongoDB's implicit ANY-match by mapping over
        # the array and applying $in on the subfield.
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        inner_lhs_mql = lhs_mql["$ifNull"][0]["$map"]["in"]
        values = process_rhs(self, compiler, connection, as_expr=True)
        lhs_mql["$ifNull"][0]["$map"]["in"] = connection.mongo_expr_operators[self.lookup_name](
            inner_lhs_mql, values
        )
        return {"$anyElementTrue": lhs_mql}


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldIn(EmbeddedModelArrayFieldBuiltinLookup, lookups.In):
    def get_subquery_wrapping_pipeline(self, compiler, connection, field_name, expr):
        # This pipeline is adapted from that of ArrayField, because the
        # structure of EmbeddedModelArrayField on the RHS behaves similar to
        # ArrayField.
        return [
            {"$project": {"subquery_results": expr.as_mql(compiler, connection, as_expr=True)}},
            # Use an $unwind followed by a $group to concatenate all the values
            # from the RHS subquery.
            {"$unwind": "$subquery_results"},
            # The $group stage collects values into an array using $addToSet.
            # The use of {_id: null} results in a single grouped array, but
            # because arrays from multiple documents are aggregated, the result
            # is a list of lists.
            {
                "$group": {
                    "_id": None,
                    "subquery_results": {"$addToSet": "$subquery_results"},
                }
            },
            # Workaround for https://jira.mongodb.org/browse/SERVER-114196:
            # $$NOW becomes unavailable after $unionWith, so it must be stored
            # beforehand to ensure it remains accessible later in the pipeline.
            {"$addFields": {"__now": "$$NOW"}},
            # Add a dummy document in case of an empty result.
            {"$unionWith": {"pipeline": [{"$documents": [{"subquery_results": []}]}]}},
            {"$limit": 1},
            {"$project": {field_name: "$subquery_results"}},
        ]


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldExact(EmbeddedModelArrayFieldBuiltinLookup, lookups.Exact):
    pass


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldIExact(EmbeddedModelArrayFieldBuiltinLookup, lookups.IExact):
    get_db_prep_lookup_value_is_iterable = False


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldGreaterThan(EmbeddedModelArrayFieldBuiltinLookup, lookups.GreaterThan):
    pass


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldGreaterThanOrEqual(
    EmbeddedModelArrayFieldBuiltinLookup, lookups.GreaterThanOrEqual
):
    pass


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldLessThan(EmbeddedModelArrayFieldBuiltinLookup, lookups.LessThan):
    pass


@_EmbeddedModelArrayOutputField.register_lookup
class EmbeddedModelArrayFieldLessThanOrEqual(
    EmbeddedModelArrayFieldBuiltinLookup, lookups.LessThanOrEqual
):
    pass


class EmbeddedModelArrayFieldTransform(Transform):
    field_class_name = "EmbeddedModelArrayField"
    VIRTUAL_COLUMN_ITERABLE = "item"

    def __init__(self, field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Lookups iterate over the array of embedded models. A virtual column
        # of the queried field's type represents each element.
        column_target = field.clone()
        column_name = f"${self.VIRTUAL_COLUMN_ITERABLE}.{field.column}"
        column_target.db_column = column_name
        column_target.set_attributes_from_name(column_name)
        self._field = field
        self._lhs = Col(None, column_target)
        self._sub_transform = None

    def __call__(self, this, *args, **kwargs):
        self._lhs = self._sub_transform(self._lhs, *args, **kwargs)
        return self

    def get_lookup(self, name):
        return self.output_field.get_lookup(name)

    def get_transform(self, name):
        """
        Validate that `name` is either a field of an embedded model or am
        allowed lookup on an embedded model's field.
        """
        # Once the sub-lhs is a transform, all the filters are applied over it.
        # Otherwise get the transform from the nested embedded model field.
        if transform := self._lhs.get_transform(name):
            if isinstance(transform, EmbeddedModelArrayFieldTransformFactory):
                raise ValueError("Cannot perform multiple levels of array traversal in a query.")
            self._sub_transform = transform
            return self
        output_field = self._lhs.output_field
        # The lookup must be allowed AND a valid lookup for the field.
        allowed_lookups = self.output_field.ALLOWED_LOOKUPS.intersection(
            set(output_field.get_lookups())
        )
        suggested_lookups = difflib.get_close_matches(name, allowed_lookups)
        if suggested_lookups:
            suggested_lookups = " or ".join(suggested_lookups)
            suggestion = f", perhaps you meant {suggested_lookups}?"
        else:
            suggestion = ""
        raise FieldDoesNotExist(
            f"Unsupported lookup '{name}' for "
            f"{self.field_class_name} of '{output_field.__class__.__name__}'"
            f"{suggestion}"
        )

    def as_mql_expr(self, compiler, connection):
        inner_lhs_mql = self._lhs.as_mql(compiler, connection, as_expr=True)
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        return {
            "$ifNull": [
                {
                    "$map": {
                        "input": lhs_mql,
                        "as": self.VIRTUAL_COLUMN_ITERABLE,
                        "in": inner_lhs_mql,
                    }
                },
                [],
            ]
        }

    def as_mql_path(self, compiler, connection):
        inner_lhs_mql = self._lhs.as_mql(compiler, connection).removeprefix(
            f"${self.VIRTUAL_COLUMN_ITERABLE}."
        )
        lhs_mql = process_lhs(self, compiler, connection)
        return f"{lhs_mql}.{inner_lhs_mql}"

    @property
    def output_field(self):
        return _EmbeddedModelArrayOutputField(self._lhs.output_field)

    @property
    def can_use_path(self):
        return self.is_simple_column

    @cached_property
    def is_simple_column(self):
        previous = self
        while isinstance(previous, EmbeddedModelArrayFieldTransform):
            previous = previous.lhs
        return previous.is_simple_column and self._lhs.is_simple_column


class EmbeddedModelArrayFieldTransformFactory:
    def __init__(self, field):
        self.field = field

    def __call__(self, *args, **kwargs):
        return EmbeddedModelArrayFieldTransform(self.field, *args, **kwargs)



import unittest
from collections.abc import Callable
from functools import wraps
from time import monotonic, sleep

from django.db import connection
from django.db.models import Q
from django.db.models.query import QuerySet
from django.db.utils import DatabaseError
from django.test import TransactionTestCase, skipUnlessDBFeature

from django_mongodb_backend.expressions import (
    CompoundExpression,
    SearchAutocomplete,
    SearchEquals,
    SearchExists,
    SearchGeoShape,
    SearchGeoWithin,
    SearchIn,
    SearchMoreLikeThis,
    SearchPhrase,
    SearchQueryString,
    SearchRange,
    SearchRegex,
    SearchScoreOption,
    SearchText,
    SearchVector,
    SearchWildcard,
)
from django_mongodb_backend.indexes import SearchIndex, VectorSearchIndex

from .models import Article, Location, Writer


def _delayed_assertion(timeout: float = 4, interval: float = 0.5):
    def decorator(assert_func):
        @wraps(assert_func)
        def wrapper(self, fetch, *args, **kwargs):
            start = monotonic()
            if not isinstance(fetch, (Callable, QuerySet)):
                raise ValueError(
                    "The first argument to a delayed assertion must be a QuerySet or a callable "
                    "that returns the value to be asserted."
                )
            if isinstance(fetch, QuerySet):
                fetch = fetch.all
            while True:
                try:
                    return assert_func(self, fetch(), *args, **kwargs)
                except (AssertionError, DatabaseError):
                    if monotonic() - start > timeout:
                        raise
                    sleep(interval)

        wrapper.__name__ = f"delayed{assert_func.__name__.title()}"
        return wrapper

    return decorator


@skipUnlessDBFeature("supports_search")
class SearchUtilsMixin(TransactionTestCase):
    available_apps = None

    """
    These assertions include a small delay to account for MongoDB Search's
    eventual consistency and indexing latency. Data inserted into MongoDB is not
    immediately available for $search queries because search indexes are
    updated asynchronously via change streams. While this is usually fast, delays
    can occur due to replication lag, system load, index complexity, or a high
    number of search indexes.
    """
    assertCountEqual = _delayed_assertion(timeout=2)(TransactionTestCase.assertCountEqual)
    assertListEqual = _delayed_assertion(timeout=2)(TransactionTestCase.assertListEqual)
    assertQuerySetEqual = _delayed_assertion(timeout=2)(TransactionTestCase.assertQuerySetEqual)

    @classmethod
    def create_search_index(cls, model, index_name, field_mappings, index_cls=SearchIndex):
        idx = index_cls(field_mappings=field_mappings, name=index_name)
        with connection.schema_editor() as editor:
            editor.add_index(model, idx)

        def drop_index():
            with connection.schema_editor() as editor:
                editor.remove_index(model, idx)

        cls.addClassCleanup(drop_index)


class SearchEqualsTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "equals_headline_index",
            {"headline": {"type": "token"}, "number": {"type": "number"}},
        )

    def setUp(self):
        self.article = Article.objects.create(headline="cross", number=1, body="body")
        Article.objects.create(headline="other thing", number=2, body="body")

    def test_search_equals(self):
        qs = Article.objects.annotate(score=SearchEquals(path="headline", value="cross"))
        self.assertCountEqual(qs, [self.article])

    def test_boost_score(self):
        boost_score = SearchScoreOption({"boost": {"value": 3}})
        qs = Article.objects.annotate(
            score=SearchEquals(path="headline", value="cross", score=boost_score)
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertGreaterEqual(scored.score, 3.0)

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchEquals(path="headline", value="cross", score=constant_score)
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_function_score(self):
        function_score = SearchScoreOption(
            {
                "function": {
                    "path": {
                        "value": "number",
                        "undefined": 0,
                    },
                }
            }
        )

        qs = Article.objects.annotate(
            score=SearchEquals(path="headline", value="cross", score=function_score)
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 1.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchEquals(path="headline", value="cross", score=score)
        self.assertEqual(str(se), f"<SearchEquals(path='headline', value='cross', score={score})>")


class SearchAutocompleteTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "autocomplete_headline_index",
            {
                "headline": {
                    "type": "autocomplete",
                    "analyzer": "lucene.standard",
                    "tokenization": "edgeGram",
                    "minGrams": 3,
                    "maxGrams": 5,
                    "foldDiacritics": False,
                },
                "writer": {
                    "type": "document",
                    "fields": {
                        "name": {
                            "type": "autocomplete",
                            "analyzer": "lucene.standard",
                            "tokenization": "edgeGram",
                            "minGrams": 3,
                            "maxGrams": 5,
                            "foldDiacritics": False,
                        }
                    },
                },
            },
        )

    def setUp(self):
        self.article = Article.objects.create(
            headline="crossing and something",
            number=2,
            body="river",
            writer=Writer(name="Joselina A. Ramirez"),
        )
        Article.objects.create(headline="Some random text", number=3, body="river")

    def test_search_autocomplete(self):
        qs = Article.objects.annotate(
            score=SearchAutocomplete(
                path="headline",
                query="crossing",
                token_order="sequential",  # noqa: S106
                fuzzy={"maxEdits": 2},
            )
        )
        self.assertCountEqual(qs, [self.article])

    def test_search_autocomplete_embedded_model(self):
        qs = Article.objects.annotate(
            score=SearchAutocomplete(path="writer__name", query="Joselina")
        )
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchAutocomplete(
                path="headline",
                query="crossing",
                token_order="sequential",  # noqa: S106
                fuzzy={"maxEdits": 2},
                score=constant_score,
            )
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchAutocomplete(path="writer__name", query="Joselina", score=score)
        self.assertEqual(
            str(se),
            "<SearchAutocomplete(path='writer__name', query='Joselina', fuzzy=None,"
            f" token_order=None, score={score})>",
        )


class SearchExistsTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "exists_body_index",
            {"body": {"type": "token"}},
        )

    def setUp(self):
        self.article = Article.objects.create(headline="ignored", number=3, body="something")

    def test_search_exists(self):
        qs = Article.objects.annotate(score=SearchExists(path="body"))
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(score=SearchExists(path="body", score=constant_score))
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchExists(path="body", score=score)
        self.assertEqual(str(se), f"<SearchExists(path='body', score={score})>")


class SearchInTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "in_headline_index",
            {"headline": {"type": "token"}},
        )

    def setUp(self):
        self.article = Article.objects.create(headline="cross", number=1, body="a")
        Article.objects.create(headline="road", number=2, body="b")

    def test_search_in(self):
        qs = Article.objects.annotate(score=SearchIn(path="headline", value=["cross", "river"]))
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchIn(path="headline", value=["cross", "river"], score=constant_score)
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchIn(path="headline", value=["cross", "river"], score=score)
        self.assertEqual(
            str(se), f"<SearchIn(path='headline', value=('cross', 'river'), score={score})>"
        )


class SearchPhraseTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "phrase_body_index",
            {"body": {"type": "string"}},
        )

    def setUp(self):
        self.article = Article.objects.create(
            headline="irrelevant", number=1, body="the quick brown fox"
        )
        Article.objects.create(headline="cheetah", number=2, body="fastest animal")

    def test_search_phrase(self):
        qs = Article.objects.annotate(score=SearchPhrase(path="body", query="quick brown", slop=3))
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchPhrase(path="body", query="quick brown", score=constant_score)
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchPhrase(path="body", query="quick brown", score=score)
        self.assertEqual(
            str(se),
            "<SearchPhrase(path='body', query='quick brown', slop=None, "
            f"synonyms=None, score={score})>",
        )


@skipUnlessDBFeature("supports_search")
class SearchQueryStringTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "query_string_index",
            {
                "headline": {"type": "string"},
                "body": {"type": "string"},
            },
        )

    def setUp(self):
        self.mars_mission = Article.objects.create(
            headline="space exploration",
            body="NASA launches a new mission to Mars",
            number=1,
        )
        self.exoplanet = Article.objects.create(
            headline="space exploration",
            body="Astronomers discover exoplanets orbiting distant stars",
            number=2,
        )
        Article.objects.create(
            headline="other news",
            body="Local team wins championship",
            number=3,
        )

    def test_search_query_string_basic(self):
        qs = Article.objects.annotate(
            score=SearchQueryString(path="headline", query="space AND body:Mars")
        )
        self.assertCountEqual(qs, [self.mars_mission])

    def test_search_query_string_with_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchQueryString(
                path="headline",
                query="space AND body:exoplanets",
                score=constant_score,
            )
        )
        self.assertCountEqual(qs, [self.exoplanet])
        self.assertAlmostEqual(qs.first().score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchQueryString(path="body", query="space AND body:exoplanets", score=score)
        self.assertEqual(
            str(se),
            f"<SearchQueryString(path='body', query='space AND body:exoplanets', score={score})>",
        )


class SearchRangeTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "range_number_index",
            {"number": {"type": "number"}},
        )
        Article.objects.create(headline="x", number=5, body="z")

    def setUp(self):
        self.number20 = Article.objects.create(headline="y", number=20, body="z")

    def test_search_range(self):
        qs = Article.objects.annotate(score=SearchRange(path="number", gte=10, lt=30))
        self.assertCountEqual(qs, [self.number20])
        qs = Article.objects.annotate(score=SearchRange(path="number", gt=20, lte=30))
        self.assertCountEqual(qs, [])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchRange(path="number", gte=10, lt=30, score=constant_score)
        )
        self.assertCountEqual(qs, [self.number20])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchRange(path="number", gte=10, lt=30, score=score)
        self.assertEqual(
            str(se),
            f"<SearchRange(path='number', lt=30, lte=None, gt=None, gte=10, score={score})>",
        )


class SearchRegexTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "regex_headline_index",
            {"headline": {"type": "string", "analyzer": "lucene.keyword"}},
        )

    def setUp(self):
        self.article = Article.objects.create(headline="hello world", number=1, body="abc")
        Article.objects.create(headline="hola mundo", number=2, body="abc")

    def test_search_regex(self):
        qs = Article.objects.annotate(
            score=SearchRegex(path="headline", query="hello.*", allow_analyzed_field=True)
        )
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchRegex(
                path="headline", query="hello.*", allow_analyzed_field=True, score=constant_score
            )
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchRegex(path="headline", query="hello.*", allow_analyzed_field=True, score=score)
        self.assertEqual(
            str(se),
            "<SearchRegex(path='headline', query='hello.*', "
            f"allow_analyzed_field=True, score={score})>",
        )


class SearchTextTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "text_body_index",
            {"body": {"type": "string"}},
        )

    def setUp(self):
        self.article = Article.objects.create(
            headline="ignored", number=1, body="The lazy dog sleeps"
        )
        Article.objects.create(headline="ignored", number=2, body="The sleepy bear")

    def test_search_text(self):
        qs = Article.objects.annotate(score=SearchText(path="body", query="lazy"))
        self.assertCountEqual(qs, [self.article])

    def test_search_lookup(self):
        qs = Article.objects.filter(body__search="lazy")
        self.assertCountEqual(qs, [self.article])

    def test_search_text_with_fuzzy_and_criteria(self):
        qs = Article.objects.annotate(
            score=SearchText(
                path="body", query="lazzy", fuzzy={"maxEdits": 2}, match_criteria="all"
            )
        )
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchText(
                path="body",
                query="lazzy",
                fuzzy={"maxEdits": 2},
                match_criteria="all",
                score=constant_score,
            )
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchText(
            path="body",
            query="lazzy",
            fuzzy={"maxEdits": 2},
            match_criteria="all",
            score=score,
        )
        self.assertEqual(
            str(se),
            "<SearchText(path='body', query='lazzy', fuzzy=(('maxEdits', 2),), "
            f"match_criteria='all', synonyms=None, score={score})>",
        )


class SearchWildcardTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "wildcard_headline_index",
            {"headline": {"type": "string", "analyzer": "lucene.keyword"}},
        )

    def setUp(self):
        self.article = Article.objects.create(headline="dark-knight", number=1, body="")
        Article.objects.create(headline="batman", number=2, body="")

    def test_search_wildcard(self):
        qs = Article.objects.annotate(
            score=SearchWildcard(path="headline", query="dark-*", allow_analyzed_field=False)
        )
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchWildcard(path="headline", query="dark-*", score=constant_score)
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        se = SearchWildcard(path="headline", query="dark-*", score=score)
        self.assertEqual(
            str(se),
            "<SearchWildcard(path='headline', query='dark-*', "
            f"allow_analyzed_field=None, score={score})>",
        )


class SearchGeoShapeTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "geoshape_location_index",
            {"location": {"type": "geo", "indexShapes": True}},
        )

    def setUp(self):
        self.article = Article.objects.create(
            headline="any", number=1, body="", location=Location(type="Point", coordinates=[40, 5])
        )
        Article.objects.create(
            headline="any",
            number=2,
            body="",
            location=Location(type="Point", coordinates=[400, 50]),
        )

    def test_search_geo_shape(self):
        polygon = {
            "type": "Polygon",
            "coordinates": [[[30, 0], [50, 0], [50, 10], [30, 10], [30, 0]]],
        }
        qs = Article.objects.annotate(
            score=SearchGeoShape(path="location", relation="within", geometry=polygon)
        )
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        polygon = {
            "type": "Polygon",
            "coordinates": [[[30, 0], [50, 0], [50, 10], [30, 10], [30, 0]]],
        }
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchGeoShape(
                path="location", relation="within", geometry=polygon, score=constant_score
            )
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        polygon = {
            "type": "Polygon",
            "coordinates": [[[30, 0], [50, 0], [50, 10], [30, 10], [30, 0]]],
        }
        se = SearchGeoShape(path="location", relation="within", geometry=polygon, score=score)
        self.assertEqual(
            str(se),
            "<SearchGeoShape(path='location', relation='within', geometry=(('type', 'Polygon'), "
            "('coordinates', (((30, 0), (50, 0), (50, 10), (30, 10), (30, 0)),))), "
            f"score={score})>",
        )


class SearchGeoWithinTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "geowithin_location_index",
            {"location": {"type": "geo"}},
        )

    def setUp(self):
        self.article = Article.objects.create(
            headline="geo", number=2, body="", location=Location(type="Point", coordinates=[40, 5])
        )
        Article.objects.create(
            headline="geo2",
            number=3,
            body="",
            location=Location(type="Point", coordinates=[-40, -5]),
        )

    def test_search_geo_within(self):
        polygon = {
            "type": "Polygon",
            "coordinates": [[[30, 0], [50, 0], [50, 10], [30, 10], [30, 0]]],
        }
        qs = Article.objects.annotate(
            score=SearchGeoWithin(
                path="location",
                kind="geometry",
                geometry=polygon,
            )
        )
        self.assertCountEqual(qs, [self.article])

    def test_constant_score(self):
        polygon = {
            "type": "Polygon",
            "coordinates": [[[30, 0], [50, 0], [50, 10], [30, 10], [30, 0]]],
        }
        constant_score = SearchScoreOption({"constant": {"value": 10}})
        qs = Article.objects.annotate(
            score=SearchGeoWithin(
                path="location",
                kind="geometry",
                geometry=polygon,
                score=constant_score,
            )
        )
        self.assertCountEqual(qs, [self.article])
        scored = qs.first()
        self.assertAlmostEqual(scored.score, 10.0, places=2)

    def test_str_returns_expected_format(self):
        score = SearchScoreOption({"constant": {"value": 10}})
        polygon = {
            "type": "Polygon",
            "coordinates": [[[30, 0], [50, 0], [50, 10], [30, 10], [30, 0]]],
        }
        se = SearchGeoWithin(
            path="location",
            kind="geometry",
            geometry=polygon,
            score=score,
        )
        self.assertEqual(
            str(se),
            "<SearchGeoWithin(path='location', kind='geometry', geometry=(('type', 'Polygon'), "
            "('coordinates', (((30, 0), (50, 0), (50, 10), (30, 10), (30, 0)),))), "
            f"score={score})>",
        )


@unittest.expectedFailure
class SearchMoreLikeThisTests(SearchUtilsMixin):
    """Expected failure: Cannot find a matching document in search index."""

    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "mlt_index",
            {"body": {"type": "string"}, "headline": {"type": "string"}},
        )
        cls.article1 = Article.objects.create(
            headline="Space exploration", number=1, body="Webb telescope"
        )
        cls.article2 = Article.objects.create(
            headline="The commodities fall",
            number=2,
            body="Commodities dropped sharply due to inflation concerns",
        )
        Article.objects.create(
            headline="irrelevant",
            number=3,
            body="This is a completely unrelated article about cooking",
        )

    def test_search_more_like_this(self):
        like_docs = [
            {"headline": self.article1.headline, "body": self.article1.body},
            {"headline": self.article2.headline, "body": self.article2.body},
        ]
        qs = Article.objects.annotate(score=SearchMoreLikeThis(documents=like_docs)).order_by(
            "score"
        )
        self.assertQuerySetEqual(qs, [self.article1, self.article2], lambda a: a.headline)


class CompoundSearchTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        cls.create_search_index(
            Article,
            "compound_index",
            {
                "headline": [{"type": "token"}, {"type": "string"}],
                "body": {"type": "string"},
                "number": {"type": "number"},
            },
        )

    def setUp(self):
        self.mars_mission = Article.objects.create(
            number=1,
            headline="space exploration",
            body="NASA launches a new mission to Mars, aiming to study surface geology",
        )

        self.exoplanet = Article.objects.create(
            number=2,
            headline="space exploration",
            body="Astronomers discover exoplanets orbiting distant stars using Webb telescope",
        )

        self.icy_moons = Article.objects.create(
            number=3,
            headline="space exploration",
            body="ESA prepares a robotic expedition to explore the icy moons of Jupiter",
        )

        self.comodities_drop = Article.objects.create(
            number=4,
            headline="astronomy news",
            body="Commodities dropped sharply due to inflation concerns",
        )

    def test_expression(self):
        must_expr = SearchEquals(path="headline", value="space exploration")
        must_not_expr = SearchPhrase(path="body", query="icy moons")
        should_expr = SearchPhrase(path="body", query="exoplanets")

        compound = CompoundExpression(
            must=[must_expr or should_expr],
            must_not=[must_not_expr],
            should=[should_expr],
            minimum_should_match=1,
        )

        qs = Article.objects.annotate(score=compound).order_by("score")
        self.assertCountEqual(qs, [self.exoplanet])

    def test_operations(self):
        expr = SearchEquals(path="headline", value="space exploration") & ~SearchEquals(
            path="number", value=3
        )
        qs = Article.objects.annotate(score=expr)
        self.assertCountEqual(qs, [self.mars_mission, self.exoplanet])

    def test_mixed_scores(self):
        boost_score = SearchScoreOption({"boost": {"value": 5}})
        constant_score = SearchScoreOption({"constant": {"value": 20}})
        function_score = SearchScoreOption(
            {"function": {"path": {"value": "number", "undefined": 0}}}
        )

        must_expr = SearchEquals(path="headline", value="space exploration", score=boost_score)
        should_expr = SearchPhrase(path="body", query="exoplanets", score=constant_score)
        must_not_expr = SearchPhrase(path="body", query="icy moons", score=function_score)
        filter_ = SearchRange(path="number", gte=1, lt=4)

        compound = CompoundExpression(
            must=[must_expr],
            must_not=[must_not_expr],
            should=[should_expr],
            filter=[filter_],
        )
        qs = Article.objects.annotate(score=compound).order_by("-score")
        self.assertListEqual(lambda: list(qs.all()), [self.exoplanet, self.mars_mission])
        # Exoplanet should rank first because of the constant 20 bump.
        self.assertEqual(qs.first(), self.exoplanet)

    def test_operationss_with_function_score(self):
        function_score = SearchScoreOption(
            {"function": {"path": {"value": "number", "undefined": 0}}}
        )

        expr = SearchEquals(
            path="headline",
            value="space exploration",
            score=function_score,
        ) & ~SearchEquals(path="number", value=3)

        qs = Article.objects.annotate(score=expr).order_by("-score")
        self.assertListEqual(lambda: list(qs.all()), [self.exoplanet, self.mars_mission])
        # Returns mars_mission (score≈1) and exoplanet (score≈2) then; exoplanet first.
        self.assertEqual(qs.first(), self.exoplanet)

    def test_multiple_search(self):
        msg = (
            "Only one $search operation is allowed per query. Received 2 search expressions. "
            "To combine multiple search expressions, use either a CompoundExpression for "
            "fine-grained control or CombinedSearchExpression for simple logical combinations."
        )
        with self.assertRaisesMessage(ValueError, msg):
            Article.objects.annotate(
                score1=SearchEquals(path="headline", value="space exploration"),
                score2=~SearchEquals(path="number", value=3),
            ).order_by("score1", "score2").first()

        with self.assertRaisesMessage(ValueError, msg):
            Article.objects.filter(
                Q(headline__search="space exploration"), Q(headline__search="space exploration 2")
            ).first()

    def test_multiple_type_search(self):
        msg = (
            "Cannot combine a `$vectorSearch` with a `$search` operator. "
            "If you need to combine them, consider "
            "restructuring your query logic or running them as separate queries."
        )
        with self.assertRaisesMessage(ValueError, msg):
            Article.objects.annotate(
                score1=SearchEquals(path="headline", value="space exploration"),
                score2=SearchVector(
                    path="headline",
                    query_vector=[1, 2, 3],
                    num_candidates=5,
                    limit=2,
                ),
            ).order_by("score1", "score2").first()

    def test_multiple_vector_search(self):
        msg = (
            "Cannot combine two `$vectorSearch` operator. If you need to combine them, "
            "consider restructuring your query logic or running them as separate queries."
        )
        with self.assertRaisesMessage(ValueError, msg):
            Article.objects.annotate(
                score1=SearchVector(
                    path="headline",
                    query_vector=[1, 2, 3],
                    num_candidates=5,
                    limit=2,
                ),
                score2=SearchVector(
                    path="headline",
                    query_vector=[1, 2, 4],
                    num_candidates=5,
                    limit=2,
                ),
            ).order_by("score1", "score2").first()

    def test_search_and_filter(self):
        qs = Article.objects.filter(headline__search="space exploration", number__gt=2)
        self.assertCountEqual(qs, [self.icy_moons])

    def test_str_returns_expected_format(self):
        must_expr = SearchEquals(path="headline", value="space exploration")
        must_not_expr = SearchPhrase(path="body", query="icy moons")
        should_expr = SearchPhrase(path="body", query="exoplanets")

        se = CompoundExpression(
            must=[must_expr or should_expr],
            must_not=[must_not_expr],
            should=[should_expr],
            minimum_should_match=1,
        )
        self.assertEqual(
            str(se),
            "<CompoundExpression(must=(<SearchEquals(path='headline', value='space exploration', "
            "score=None)>,), must_not=(<SearchPhrase(path='body', query='icy moons', slop=None, "
            "synonyms=None, score=None)>,), should=(<SearchPhrase(path='body', "
            "query='exoplanets', slop=None, synonyms=None, score=None)>,), "
            "filter=None, score=None, minimum_should_match=1)>",
        )


class SearchVectorTests(SearchUtilsMixin):
    @classmethod
    def setUpClass(cls):
        model = Article
        idx = VectorSearchIndex(
            fields=["plot_embedding", "number"],
            name="vector_index",
            similarities="cosine",
        )
        with connection.schema_editor() as editor:
            editor.add_index(model, idx)

        def drop_index():
            with connection.schema_editor() as editor:
                editor.remove_index(model, idx)

        cls.addClassCleanup(drop_index)

    def setUp(self):
        self.mars = Article.objects.create(
            headline="Mars landing",
            number=1,
            body="The rover has landed on Mars",
            plot_embedding=[0.1, 0.2, 0.3],
        )
        self.cooking = Article.objects.create(
            headline="Cooking tips",
            number=2,
            body="This article is about pasta",
            plot_embedding=[0.9, 0.8, 0.7],
        )
        Article.objects.create(
            headline="Local team wins championship",
            number=3,
            body="This article is about sports",
            plot_embedding=[-0.1, 0.7, 0.7],
        )

    def test_vector_search(self):
        vector_query = [0.1, 0.2, 0.3]
        expr = SearchVector(
            path="plot_embedding",
            query_vector=vector_query,
            num_candidates=5,
            limit=3,
            filter={"number": {"$lt": 3}},
            exact=False,
        )
        qs = Article.objects.annotate(score=expr).order_by("-score")
        self.assertCountEqual(qs, [self.mars, self.cooking])

    def test_str_returns_expected_format(self):
        vector_query = [0.1, 0.2, 0.3]
        se = SearchVector(
            path="plot_embedding",
            query_vector=vector_query,
            num_candidates=5,
            limit=2,
        )
        self.assertEqual(
            str(se),
            "<SearchVector(path='plot_embedding', query_vector=(0.1, 0.2, 0.3), limit=2, "
            "num_candidates=5, exact=None, filter=None)>",
        )



from django.db import NotSupportedError
from django.db.models.aggregates import (
    Aggregate,
    Count,
    StdDev,
    StringAgg,
    Variance,
)
from django.db.models.expressions import Case, Value, When
from django.db.models.functions.comparison import Coalesce
from django.db.models.lookups import IsNull
from django.db.models.sql.where import WhereNode

from django_mongodb_backend.expressions import Remove

# Aggregates whose MongoDB aggregation name differ from Aggregate.function.lower().
MONGO_AGGREGATIONS = {Count: "sum"}


def aggregate(self, compiler, connection, operator=None, resolve_inner_expression=False):
    agg_expression, *_ = self.get_source_expressions()
    lhs_mql = None
    if self.filter is not None:
        try:
            lhs_mql = self.filter.as_mql(compiler, connection, as_expr=True)
        except NotSupportedError:
            # Generate a CASE statement for this AggregateFilter.
            agg_expression = Case(
                When(self.filter.condition, then=agg_expression),
                # Skip rows that don't meet the criteria.
                default=Remove(),
            )
    if lhs_mql is None:
        lhs_mql = agg_expression.as_mql(compiler, connection, as_expr=True)
    if resolve_inner_expression:
        return lhs_mql
    operator = operator or MONGO_AGGREGATIONS.get(self.__class__, self.function.lower())
    return {f"${operator}": lhs_mql}


def count(self, compiler, connection, resolve_inner_expression=False):
    """
    When resolve_inner_expression=True, return the MQL that resolves as a
    value. This is used to count different elements, so the inner values are
    returned to be pushed into a set.
    """
    agg_expression, *_ = self.get_source_expressions()
    if not self.distinct or resolve_inner_expression:
        lhs_mql = None
        conditions = [IsNull(agg_expression, False)]
        if self.filter:
            try:
                lhs_mql = self.filter.as_mql(compiler, connection, as_expr=True)
            except NotSupportedError:
                # Generate a CASE statement for this AggregateFilter.
                conditions.append(self.filter.condition)
                condition = When(
                    WhereNode(conditions),
                    then=agg_expression if self.distinct else Value(1),
                )
                inner_expression = Case(condition, default=Remove())
        else:
            inner_expression = Case(
                When(WhereNode(conditions), then=agg_expression if self.distinct else Value(1)),
                # Skip rows that don't meet the criteria.
                default=Remove(),
            )
        if lhs_mql is None:
            lhs_mql = inner_expression.as_mql(compiler, connection, as_expr=True)
        if resolve_inner_expression:
            return lhs_mql
        return {"$sum": lhs_mql}
    # Normalize empty documents (introduced by aggregation wrapping) to an
    # empty set fallback.
    agg_expression = Coalesce(agg_expression, [])
    # If distinct=True or resolve_inner_expression=False, sum the size of the
    # set.
    return {"$size": agg_expression.as_mql(compiler, connection, as_expr=True)}


def stddev_variance(self, compiler, connection):
    if self.function.endswith("_SAMP"):
        operator = "stdDevSamp"
    elif self.function.endswith("_POP"):
        operator = "stdDevPop"
    return aggregate(self, compiler, connection, operator=operator)


def string_agg(self, compiler, connection, resolve_inner_expression=False):  # noqa: ARG001
    raise NotSupportedError("StringAgg is not supported.")


def register_aggregates():
    Aggregate.as_mql_expr = aggregate
    Count.as_mql_expr = count
    StdDev.as_mql_expr = stddev_variance
    StringAgg.as_mql_expr = string_agg
    Variance.as_mql_expr = stddev_variance



from django.contrib.gis import geos
from django.contrib.gis.db import models
from django.contrib.gis.db.backends.base.operations import BaseSpatialOperations
from django.contrib.gis.measure import Distance

from .adapter import Adapter
from .operators import (
    Contains,
    Disjoint,
    DistanceGT,
    DistanceLTE,
    DWithin,
    Intersects,
    Within,
)


class GISOperations(BaseSpatialOperations):
    Adapter = Adapter

    disallowed_aggregates = (
        models.Collect,
        models.Extent,
        models.Extent3D,
        models.MakeLine,
        models.Union,
    )

    gis_operators = {
        "contains": Contains(),
        "disjoint": Disjoint(),
        "distance_gt": DistanceGT(),
        "distance_lte": DistanceLTE(),
        "dwithin": DWithin(),
        "intersects": Intersects(),
        "within": Within(),
    }

    unsupported_functions = {
        "Area",
        "AsGeoJSON",
        "AsGML",
        "AsKML",
        "AsSVG",
        "AsWKB",
        "AsWKT",
        "Azimuth",
        "BoundingCircle",
        "Centroid",
        "ClosestPoint",
        "Difference",
        "Distance",
        "Envelope",
        "ForcePolygonCW",
        "FromWKB",
        "FromWKT",
        "GeoHash",
        "GeometryDistance",
        "GeometryType",
        "Intersection",
        "IsEmpty",
        "IsValid",
        "Length",
        "LineLocatePoint",
        "MakeValid",
        "MemSize",
        "NumGeometries",
        "NumPoints",
        "Perimeter",
        "PointOnSurface",
        "Reverse",
        "Rotate",
        "Scale",
        "SnapToGrid",
        "SymDifference",
        "Transform",
        "Translate",
        "Union",
    }

    def geo_db_type(self, f):
        return "object"

    def get_geometry_converter(self, expression):
        srid = expression.output_field.srid

        def converter(value, expression, connection):  # noqa: ARG001
            if value is None:
                return None

            geom_class = getattr(geos, value["type"])
            if geom_class.__name__ == "GeometryCollection":
                return geom_class(
                    [
                        getattr(geos, v["type"])(*v["coordinates"], srid=srid)
                        for v in value["geometries"]
                    ],
                    srid=srid,
                )
            if issubclass(geom_class, geos.GeometryCollection):
                sub_geom_class = geom_class._allowed
                # MultiLineString allows both LineString and LinearRing but should be
                # initialized with LineString.
                if isinstance(sub_geom_class, tuple):
                    sub_geom_class = sub_geom_class[0]
                return geom_class(
                    [
                        sub_geom_class(*value["coordinates"][x])
                        for x in range(len(value["coordinates"]))
                    ],
                    srid=srid,
                )
            return geom_class(*value["coordinates"], srid=srid)

        return converter

    def get_distance(self, f, value, lookup_type):
        value = value[0]
        if isinstance(value, Distance):
            raise ValueError("Only numeric values of degree units are allowed on dwithin queries.")
        return [value]



from bson import ObjectId
from django.test import SimpleTestCase
from django.test.utils import override_settings
from django.urls import Resolver404, resolve, reverse


@override_settings(ROOT_URLCONF="urlpatterns_.urls")
class ObjectIdConverterTests(SimpleTestCase):
    def test_resolve(self):
        match = resolve("/articles/69868bc49b827bee857500c2/")
        self.assertEqual(match.url_name, "article-detail")
        self.assertEqual(match.args, ())
        self.assertEqual(match.kwargs, {"pk": ObjectId("69868bc49b827bee857500c2")})

    def test_reverse(self):
        url = reverse("article-detail", kwargs={"pk": ObjectId("69868bc49b827bee857500c2")})
        self.assertEqual(url, "/articles/69868bc49b827bee857500c2/")

    def test_invalid_character(self):
        # Correct length, but "z" not a valid character for ObjectId.
        with self.assertRaises(Resolver404):
            resolve("/articles/69868bc49b827bee857500cz/")

    def test_too_few_characters(self):
        # ObjectId must be 24 characters, not 23.
        with self.assertRaises(Resolver404):
            resolve("/articles/69868bc49b827bee857500c/")

    def test_too_many_characters(self):
        # ObjectId must be 24 characters, not 25.
        with self.assertRaises(Resolver404):
            resolve("/articles/69868bc49b827bee857500c21/")



from django.db import connection
from django.test import TestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.constraints import EmbeddedFieldUniqueConstraint

from .models import DataHolder, Movie, Owner, Person, Store


@isolate_apps("constraints_")
class EmbeddedFieldUniqueConstraintTests(TestCase):
    def test_valid_embedded_model_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["data.integer"],
        )
        errors = constraint.check(DataHolder, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_embedded_model_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=[
                "data.field_does_not_exist",
                "another_non_existing",
                "data.integer",
            ],
        )
        errors = constraint.check(DataHolder, connection=connection)
        self.assertEqual(len(errors), 2)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'constraints' refers to the nonexistent field 'data.field_does_not_exist'.",
        )
        self.assertEqual(errors[1].id, "models.E012")
        self.assertEqual(
            errors[1].msg, "'constraints' refers to the nonexistent field 'another_non_existing'."
        )

    def test_valid_embedded_model_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["reviews.title"],
            nulls_distinct=False,
        )
        errors = constraint.check(Movie, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_embedded_model_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["reviews.author", "non_existing", "reviews.title"],
            nulls_distinct=False,
        )
        errors = constraint.check(Movie, connection=connection)
        self.assertEqual(len(errors), 2)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'constraints' refers to the nonexistent field 'reviews.author'.",
        )
        self.assertEqual(errors[1].id, "models.E012")
        self.assertEqual(
            errors[1].msg, "'constraints' refers to the nonexistent field 'non_existing'."
        )

    def test_embedded_model_array_subfield_without_nulls_distinct(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["reviews.title"],
        )
        errors = constraint.check(Movie, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.constraints.embedded_unique.E001")
        self.assertEqual(
            errors[0].msg,
            "EmbeddedFieldUniqueConstraint 'constraint_name' must have "
            "nulls_distinct=False since it references EmbeddedModelArrayField "
            "'reviews'.",
        )

    def test_valid_polymorphic_embedded_model_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["pet.name"],
        )
        errors = constraint.check(Person, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_polymorphic_embedded_model_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["pet.nonexistent"],
        )
        errors = constraint.check(Person, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'constraints' refers to the nonexistent field 'pet.nonexistent'.",
        )

    def test_valid_polymorphic_embedded_model_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["pets.name"],
            nulls_distinct=False,
        )
        errors = constraint.check(Owner, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_polymorphic_embedded_model_array_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["pets.nonexistent"],
            nulls_distinct=False,
        )
        errors = constraint.check(Owner, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'constraints' refers to the nonexistent field 'pets.nonexistent'.",
        )

    def test_polymorphic_embedded_model_array_subfield_without_nulls_distinct(self):
        constraint = EmbeddedFieldUniqueConstraint(
            name="constraint_name",
            fields=["pets.name"],
        )
        errors = constraint.check(Owner, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.constraints.embedded_unique.E001")
        self.assertEqual(
            errors[0].msg,
            "EmbeddedFieldUniqueConstraint 'constraint_name' must have "
            "nulls_distinct=False since it references PolymorphicEmbeddedModelArrayField "
            "'pets'.",
        )

    def test_valid_nested_polymorphic_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["thing.tags.name"])
        errors = constraint.check(Store, connection=connection)
        self.assertEqual(errors, [])

    def test_nonexistent_nested_polymorphic_subfield(self):
        constraint = EmbeddedFieldUniqueConstraint(name="name", fields=["thing.tags.xxx"])
        errors = constraint.check(Store, connection=connection)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "models.E012")
        self.assertEqual(
            errors[0].msg,
            "'constraints' refers to the nonexistent field 'thing.tags.xxx'.",
        )



from django.db import models

from django_mongodb_backend.fields import (
    ArrayField,
    EmbeddedModelField,
)
from django_mongodb_backend.models import EmbeddedModel


class Writer(EmbeddedModel):
    name = models.CharField(max_length=10)


class Location(EmbeddedModel):
    type = models.CharField(default="Point", max_length=50)
    coordinates = ArrayField(models.FloatField(), max_size=2)


class Article(models.Model):
    headline = models.CharField(max_length=100)
    number = models.IntegerField()
    body = models.TextField()
    location = EmbeddedModelField(Location, null=True)
    plot_embedding = ArrayField(models.FloatField(), size=3, null=True)
    writer = EmbeddedModelField(Writer, null=True)



"""Literals that MongoDB intreprets as expressions are escaped."""

from operator import attrgetter

from django.db.models import Value
from django.test import TestCase

from django_mongodb_backend.test import MongoTestCaseMixin

from .models import Author, Blob


class ModelCreationTests(MongoTestCaseMixin, TestCase):
    def test_dollar_prefixed_string(self):
        # No escaping is needed because MongoDB's insert doesn't treat
        # dollar-prefixed strings as expressions.
        with self.assertNumQueries(1) as ctx:
            obj = Author.objects.create(name="$foobar")
        obj.refresh_from_db()
        self.assertEqual(obj.name, "$foobar")
        self.assertInsertQuery(
            ctx.captured_queries[0]["sql"], "basic__author", [{"name": "$foobar"}]
        )


class ModelUpdateTests(MongoTestCaseMixin, TestCase):
    """
    $-prefixed strings and dict/tuples that could be interpreted as expressions
    are escaped in the queries that update model instances.
    """

    def test_dollar_prefixed_string(self):
        obj = Author.objects.create(name="foobar")
        obj.name = "$updated"
        with self.assertNumQueries(1) as ctx:
            obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.name, "$updated")
        self.assertUpdateOneQuery(
            ctx.captured_queries[0]["sql"],
            "basic__author",
            {"_id": obj.id},
            [{"$set": {"name": {"$literal": "$updated"}}}],
        )

    def test_dollar_prefixed_value(self):
        obj = Author.objects.create(name="foobar")
        obj.name = Value("$updated")
        with self.assertNumQueries(1) as ctx:
            obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.name, "$updated")
        self.assertUpdateOneQuery(
            ctx.captured_queries[0]["sql"],
            "basic__author",
            {"_id": obj.id},
            [{"$set": {"name": {"$literal": "$updated"}}}],
        )

    def test_dict(self):
        obj = Blob.objects.create(name="foobar")
        obj.data = {"$concat": ["$name", "-", "$name"]}
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.data, {"$concat": ["$name", "-", "$name"]})

    def test_dict_value(self):
        obj = Blob.objects.create(name="foobar", data={})
        obj.data = Value({"$concat": ["$name", "-", "$name"]})
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.data, {"$concat": ["$name", "-", "$name"]})

    def test_tuple(self):
        obj = Blob.objects.create(name="foobar")
        obj.data = ("$name", "-", "$name")
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.data, ["$name", "-", "$name"])

    def test_tuple_value(self):
        obj = Blob.objects.create(name="foobar")
        obj.data = Value(("$name", "-", "$name"))
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.data, ["$name", "-", "$name"])


class AnnotationTests(MongoTestCaseMixin, TestCase):
    def test_dollar_prefixed_value(self):
        """Value() escapes dollar prefixed strings."""
        Author.objects.create(name="Gustavo")
        with self.assertNumQueries(1) as ctx:
            qs = list(Author.objects.annotate(a_value=Value("$name")))
        self.assertQuerySetEqual(qs, ["$name"], attrgetter("a_value"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "basic__author",
            [{"$project": {"a_value": {"$literal": "$name"}, "_id": 1, "name": 1}}],
        )
